import sys

from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.model import Model
import numpy as np
import pytest

from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.components.linear import Linear
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT

sys.path.append("..")  # Adds higher directory to python modules path.


def make_multispectrum(specshape, ll=None):
    shape = (20, 15)
    m, q = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))
    data = np.zeros((shape[0], shape[1], specshape.size))
    lin = Linear(specshape, 1, 1)

    for index in (np.ndindex(shape)):
        islice = np.s_[index]
        if ll is not None:
            mod = Model(specshape, components=[lin, ll])
        else:
            mod = Model(specshape, components=[lin])

        lin.parameters[0].setvalue(m[islice])
        lin.parameters[1].setvalue(q[islice])
        mod.calculate()
        data[islice] = np.random.normal(mod.data, 0.01)

    msh = MultiSpectrumshape(specshape.dispersion, specshape.offset,
                             specshape.size, shape[0], shape[1])
    s = MultiSpectrum(msh, data=data)
    return s, m, q


def make_multispectrum_convolutor(specshape):
    shape = (20, 15)
    m, q = np.meshgrid(np.arange(shape[1]) + 1, np.arange(shape[0]) + 1)
    data = np.zeros((shape[0], shape[1], specshape.size))
    lldata = np.zeros(data.shape)
    lin = Linear(specshape, 1, 1)

    gaus1 = Gaussian(specshape, A=10, centre=500, fwhm=100)

    for index in (np.ndindex(shape)):
        islice = np.s_[index]
        gaus1.parameters[2].setvalue(50 * m[islice])
        gaus1.calculate()
        lin.parameters[0].setvalue(m[islice])
        lin.parameters[1].setvalue(q[islice])
        ll = MscatterFFT(specshape, gaus1)
        mod = Model(specshape, components=[lin, ll])
        mod.calculate()
        data[islice] = np.random.normal(mod.data, 0.01)
        lldata[islice] = gaus1.data

    msh = MultiSpectrumshape(specshape.dispersion, specshape.offset,
                             specshape.size, shape[0], shape[1])
    s = MultiSpectrum(msh, data=data)
    ll = MscatterFFT(specshape, llspectrum=MultiSpectrum(msh, lldata))
    return s, m, q, ll


def test_nonlinear():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)
    gaus = Gaussian(specshape, A=1000, centre=500, fwhm=50)

    mod = Model(specshape, components=[lin, gaus])
    mod.calculate()

    s = Spectrum(specshape, data=np.random.normal(mod.data, 10))
    with pytest.raises(ValueError):
        LinearFitter(s, mod)


def test_A_matrix():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)
    # gaus = Gaussian(specshape, A=1000, centre=500, fwhm=50)

    mod = Model(specshape, components=[lin])
    mod.calculate()

    s = Spectrum(specshape, data=np.random.normal(mod.data, 10))
    fit = LinearFitter(s, mod)
    fit.calculate_A_matrix()
    assert fit.A_matrix.shape == (2048, 2)

    # the excluding does not modify the A matrix size
    # s = Spectrum(specshape, data = np.random.normal(mod.data,10))
    # s.set_exclude_region_energy(300,500)
    # fit = LinearFitter(s, mod)
    # fit.calculate_A_matrix()
    # assert fit.A_matrix.shape == (1848,2)


def test_perform_fit():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)

    mod = Model(specshape, components=[lin])
    mod.calculate()

    s = Spectrum(specshape, data=np.random.normal(mod.data, 0.01))
    lin.parameters[0].setvalue(1.)
    lin.parameters[1].setvalue(1.)

    s.set_exclude_region_energy(300, 500)
    fit = LinearFitter(s, mod)
    fit.perform_fit()
    assert np.abs(fit.coeff[0] - 2.5) < 2.5*0.1
    assert np.abs(fit.coeff[1] - 5) < 5*0.1


def test_perform_fit_convolutor():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)

    gaus1 = Gaussian(specshape, A=10, centre=500, fwhm=100)
    gaus1.calculate()
    ll = MscatterFFT(specshape, gaus1)

    mod = Model(specshape, components=[lin, ll])
    mod.calculate()

    s = Spectrum(specshape, data=np.random.normal(mod.data, 1))
    fit = LinearFitter(s, mod)
    fit.perform_fit()

    index = fit.get_param_index(lin.parameters[0])
    assert np.abs(fit.coeff[index] - 2.5) < 1e-1*2.5
    index = fit.get_param_index(lin.parameters[1])
    assert np.abs(fit.coeff[index] - 5) < 1e-1*5


def test_multifit():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)

    mod = Model(specshape, components=[lin])
    s, m, q = make_multispectrum(specshape)

    fit = LinearFitter(s, mod)
    fit.multi_fit()

    index = fit.get_param_index(lin.parameters[0])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - m) < 0.01)
    index = fit.get_param_index(lin.parameters[1])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - q) < 0.01)


def test_multifit_convolutor():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)

    gaus1 = Gaussian(specshape, A=10, centre=500, fwhm=100)
    gaus1.calculate()
    ll = MscatterFFT(specshape, gaus1)

    mod = Model(specshape, components=[lin, ll])
    s, m, q = make_multispectrum(specshape, ll=ll)

    fit = LinearFitter(s, mod)
    fit.multi_fit()

    index = fit.get_param_index(lin.parameters[0])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - m) < 0.01)
    index = fit.get_param_index(lin.parameters[1])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - q) < 0.01)


def test_multifit_multiconvolutor():
    specshape = Spectrumshape(1, 100, 2048)
    lin = Linear(specshape, 2.5, 5)

    s, m, q, ll = make_multispectrum_convolutor(specshape)
    mod = Model(specshape, components=[lin, ll])
    fit = LinearFitter(s, mod)
    fit.multi_fit()

    index = fit.get_param_index(lin.parameters[0])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - m) < 0.01)
    index = fit.get_param_index(lin.parameters[1])
    assert np.all(np.abs(fit.coeff_matrix[:, :, index] - q) < 0.1)


def main():
    test_nonlinear()
    test_A_matrix()
    test_perform_fit()
    test_perform_fit_convolutor()
    test_multifit()
    test_multifit_convolutor()
    test_multifit_multiconvolutor()


if __name__ == "__main__":
    main()



import numpy as np
from tqdm import tqdm
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape


class Splice(Operator):
    """
    This class splines multiple spectra into each other. At this point the
    spectra should still overlap.
    The splicing first calculates a new energy axis which conveys each
    spectrum added to the spectra list.
    Then each spectrum is interpolated to the new axis and the weighted
    average is calcualted with the information on the acquisition time.
    This assumes that each spectrum is acquired with the
    same current.
    It is not needed for the dispersion to be the same. When this happens it
    uses the highest dispersion (lowest value) for the interpolation.


    """
    def __init__(self, spectra, acq_times=None):
        """
        Initializes the Splice object.

        Parameters
        ----------
        spectra: list of Spectrums or MultiSpectrums
            The list containing the spectra which need to be spliced together.
            If Multispectrums are used then it is important that the xsize and
            ysize are the same.
        acq_times: list
            A list of the acquisition times. The acquistion times do not need
            to be in absolute units (such as seconds) but the ratios should be
            conserved.


        """
        self.spectra = spectra

        if isinstance(spectra[0], MultiSpectrum):
            self.check_multispectra_validity(spectra)
            self.has_multispectra = True
        else:
            self.has_multispectra = False

        self.acq_times = acq_times
        self.weights = self.acq_times

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, weights):
        if weights is None:
            self._weights = np.ones(len(self.spectra))
        else:
            self._weights = weights

    # def normalize_weights(self):
    #     print('sum of weights is :'+str(self.weights.sum()))
    #     self.weights = len(self.spectra)*self.weights/self.weights.sum()

    def check_multispectra_validity(self, spectra):
        if len(spectra) < 2:
            raise ValueError(r'Not sufficient amount of multispectra'
                             r' in the list')
        initx = spectra[0].xsize
        inity = spectra[0].ysize
        for spectrum in spectra[1:]:
            if (spectrum.xsize != initx) | (spectrum.ysize != inity):
                raise ValueError(r'The scan size of the multispectra '
                                 r'are not the same')

    def get_new_energy_axis(self):
        """
        Determines the lowest energy, highest energy and highest dispersion
        from the list of spectra provided. From this information it
        calculates a new energy axis E

        Returns
        -------
        E: 1d numpy array
            The new energy axis on which each spectrum will be interpolated.

        """

        for i in range(len(self.spectra)-1):
            if i == 0:
                offset = min(self.spectra[i].offset,
                             self.spectra[i+1].offset)
                dispersion = min(self.spectra[i].dispersion,
                                 self.spectra[i + 1].dispersion)
                end_E = max(self.spectra[i].energy_axis[-1],
                            self.spectra[i+1].energy_axis[-1])

            else:
                offset = min(offset, self.spectra[i+1].offset)
                dispersion = min(dispersion, self.spectra[i + 1].dispersion)
                end_E = max(end_E, self.spectra[i+1].energy_axis[-1])

        E = np.arange(offset, end_E, dispersion)
        return E

    def _get_spectrum_from_energy_axis(self, energy_axis):
        """
        Function which makes a spectrum of the enery axis.
        This is needed in the way the function is implemented in spectrum

        """
        sh = Spectrumshape(np.diff(energy_axis)[0], energy_axis[0],
                           energy_axis.size)
        s = Spectrum(sh)
        return s, sh

    def _get_multispectrum_from_energy_axis(self, energy_axis):
        """
        Function which makes a spectrum of the enery axis.
        This is needed in the way the function is implemented in spectrum
        """
        xsize = self.spectra[0].xsize
        ysize = self.spectra[0].ysize
        sh = MultiSpectrumshape(np.diff(energy_axis)[0], energy_axis[0],
                                energy_axis.size, xsize, ysize)
        s = MultiSpectrum(sh)
        return s, sh

    def splice_spectra(self):
        """
        Multiple spectra can be spliced together. This only works for spectra
        and not multispectra
        """
        E = self.get_new_energy_axis()
        spectrum, sshape = self._get_spectrum_from_energy_axis(E)
        ndata = self._weigth_interpolate(spectrum)
        s = Spectrum(sshape, data=ndata)
        return s

    def _weigth_interpolate(self, s):
        ndata = np.zeros((len(self.spectra), s.size))
        # weight_array = np.zeros((len(self.spectra), spectrum.size))
        for ii, spec in enumerate(self.spectra):
            intspec = spec.interp_to_other_energy_axis(s, (np.nan, np.nan))
            boolean = np.invert(np.isnan(intspec.data))
            ndata[ii, boolean] = intspec.data[boolean]/self.acq_times[ii]
            # weight_array[ii, boolean] = self.weights[ii]

        res = (ndata * self.weight_array).sum(0)/self.weight_array.sum(0)

        return res

    def _calculate_weight_array(self, s):

        weight_array = np.zeros((len(self.spectra), s.size))
        for ii, spec in enumerate(self.spectra):
            intspec = spec.interp_to_other_energy_axis(s, (np.nan, np.nan))
            boolean = np.invert(np.isnan(intspec.data))
            weight_array[ii, boolean] = self.weights[ii]

        self.weight_array = weight_array

    def splice_multispectra(self):
        E = self.get_new_energy_axis()
        multispectrum, sshape = self._get_multispectrum_from_energy_axis(E)
        self._calculate_weight_array(multispectrum)
        shape = (multispectrum.xsize, multispectrum.ysize)
        for index in tqdm(np.ndindex(shape)):
            islice = np.s_[index]
            for spec in self.spectra:
                spec.setcurrentspectrum(index)
            multispectrum.multidata[islice] =\
                self._weigth_interpolate(multispectrum)

        return multispectrum



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
import numpy as np
from pyEELSMODEL.core.parameter import Parameter


class Lorentzian(Component):
    """
        Initialises a lorentzian component.

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the lorentzian. Be careful about the exact
            implementation.

        centre: float
            The position of the peak position
        fwhm: float
            The full width half maximum of the lorentzian.

        Returns
        -------

    """

    def __init__(self, specshape, A, centre, fwhm):
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, np.inf)
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('centre', centre)
        p2.setboundaries(-np.inf, np.inf)
        p2.sethasgradient(False)
        self._addparameter(p2)

        p3 = Parameter('FWHM', fwhm)
        p3.setboundaries(0, np.inf)
        p3.sethasgradient(False)
        self._addparameter(p3)

        self._setname('Lorentzian')

    def calculate(self):
        """
        Calculates the lorentzian function
        """
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        E = self.energy_axis
        self.data = self.lorentz(E, self.parameters[0].getvalue(),
                                 self.parameters[1].getvalue(),
                                 self.parameters[2].getvalue())

    def lorentz(self, x, A, x0, dT):
        """
        Definition of the lorentzian function. Quite important to know this
        when wanting to extract relevant infromation from it such as fwhm
        or integral.
        """

        # return A*(dT**2/4)*(1/((x-x0)**2+(0.5*dT)**2))
        # return A*(1/((x-x0)**2+(0.5*dT)**2))
        return A * (0.5 * dT) ** 2 / ((x - x0) ** 2 + (0.5 * dT) ** 2)

    def getgradient(self, parameter):
        """
        Calculates the gradient with respect to the given parameter.
        Sets this has the gradient of the component.

            Parameters
        ----------
        parameter : Parameter
            The parameter to which the derivative should be calculated.

        Returns
        -------
        """
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        A = p1.getvalue()
        centre = p2.getvalue()
        FWHM = p3.getvalue()

        en = self.energy_axis
        denom = ((en - centre) ** 2 + (0.5 * FWHM) ** 2)

        if parameter == p1:
            self.gradient[0] = (FWHM ** 2 / 4) * (1 / denom)
            return self.gradient[0]
        if parameter == p2:
            self.gradient[1] = (A * FWHM ** 2 / 2) * (en - centre) / denom ** 2
            return self.gradient[1]
        if parameter == p3:
            self.gradient[2] = 0.5 * A * FWHM * (
                        1 / denom - (FWHM ** 2 / 4) * (1 / denom ** 2))
            return self.gradient[2]

        return None



from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np


class AfterGlow(Component):
    """
    Component which can add a afterglow if it is observed in the experimental
    data.
    Simple approximation which uses the same zlp for each spectrum whereas
    drift could make this different. Additionally, the position of the after
    glow needs to be indicated which would not be needed

    #todo find a generic way to handle the afterglow with any need of Epos

    """

    def __init__(self, specshape, zlp, Epos, A=1):
        """
        Initiates the Afterglow component

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        zlp : numpy array
            The zlp array which is best taken to be the average.
        Epos: energy position of the peak of the afterglow
        """
        super().__init__(specshape)
        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, 1e20)
        p1.sethasgradient(True)
        self._addparameter(p1)

        self.setdisplayname('Afterglow')
        self.Epos = Epos
        self.zlp = zlp
        self.set_fixeddata(self.zlp, self.Epos)
        # don't convolute the background it only gives problems
        # and adds no extra physics
        self._setcanconvolute(False)

        self.calculate()

    def set_fixeddata(self, zlp, Epos):
        """
        Sets the component using the zlp data and the energy where it is seen.

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        zlp : numpy array
            The zlp array which is best taken to be the average.
        Epos: energy position of the peak of the afterglow

        """
        index = self.get_energy_index(Epos)

        shift = np.argmax(zlp) - index

        zlpdata = np.roll(zlp, -shift)[:self.size]
        pad_size = self.size - zlpdata.size
        print(pad_size)

        if pad_size > 0:
            zlpdata = np.pad(zlpdata, (0, pad_size))
            print(zlpdata.size)

        self.fixeddata = zlpdata

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]

        if p1.ischanged():
            A = p1.getvalue()
            self.data = A * self.fixeddata
        self.setunchanged()  # put parameters to unchanged



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np


class FastBG2(Component):
    """
    Fast background model, fully linear and using two terms to model the
    background
    """

    def __init__(self, specshape, A1=1, A2=1, r1=2.5, r2=3.5):
        super().__init__(specshape)
        p0 = Parameter('A1', A1)
        p0.setlinear(True)
        p0.setboundaries(-np.inf, np.inf)
        p0.sethasgradient(True)
        self._addparameter(p0)

        p1 = Parameter('r1', r1, changeallowed=False)
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('A2', A2)
        p2.setlinear(True)
        p2.setboundaries(-np.inf, np.inf)
        p2.sethasgradient(True)
        self._addparameter(p2)

        p3 = Parameter('r2', r2, changeallowed=False)
        p3.sethasgradient(False)
        self._addparameter(p3)
        # don't convolute the background it only gives problems and
        # adds no extra physics
        self._setcanconvolute(False)
        self._setshifter(False)  # it is not a shifter type of component

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        p2 = self.parameters[2]
        p3 = self.parameters[3]

        if p0.ischanged() or p1.ischanged() or p2.ischanged() \
                or p3.ischanged():
            A1 = p0.getvalue()
            r1 = p1.getvalue()
            A2 = p2.getvalue()
            r2 = p3.getvalue()
            self.data = self.fast_background(A1, r1, A2, r2)
        self.setunchanged()  # put parameters to unchanged

    def fast_background(self, A1, r1, A2, r2):
        E = self.energy_axis
        Estart = E[0]
        if Estart < 1:
            Estart = 1

        mask = (E > 0)  # only meaningfull when E>0 and r>0
        # return mask*(A1 * (E/Estart) ** (-r1) + A2 * (E/Estart) ** (-r2))
        return mask * (A1 * E ** (-r1) + A2 * E ** (-r2))

    def getgradient(self, parameter):
        """calculate the analytical partial derivative wrt parameter j
        returns true if succesful, gradient is stored in component.gradient
        """
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        p2 = self.parameters[2]
        p3 = self.parameters[3]

        r1 = p1.getvalue()
        r2 = p3.getvalue()

        E = self.energy_axis
        Estart = E[0]
        if Estart < 1:
            Estart = 1
        mask = (E > 0)  # only meaningfull when E>0 and r>0
        if parameter == p0:
            self.gradient[0] = mask * E ** (-r1)
            return self.gradient[0]
        elif parameter == p2:
            self.gradient[2] = mask * E ** (-r2)
            return self.gradient[2]
        else:
            return False


class FastBG3(Component):
    """
    Fast background model, fully linear and using three terms to model the
    background
    """

    def __init__(self, specshape, A1=1, A2=1, A3=1, r1=2, r2=3, r3=4):
        super().__init__(specshape)
        p0 = Parameter('A1', A1)
        p0.setlinear(True)
        p0.setboundaries(-np.inf, np.inf)
        p0.sethasgradient(True)
        self._addparameter(p0)

        p1 = Parameter('r1', r1, changeallowed=False)
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('A2', A2)
        p2.setlinear(True)
        p2.setboundaries(-np.inf, np.inf)
        p2.sethasgradient(True)
        self._addparameter(p2)

        p3 = Parameter('r2', r2, changeallowed=False)
        p3.sethasgradient(False)
        self._addparameter(p3)

        p4 = Parameter('A3', A3)
        p4.setlinear(True)
        p4.setboundaries(-np.inf, np.inf)
        p4.sethasgradient(True)
        self._addparameter(p4)

        p5 = Parameter('r3', r3, changeallowed=False)
        p5.sethasgradient(False)
        self._addparameter(p5)
        # don't convolute the background it only gives problems
        # and adds no extra physics
        self._setcanconvolute(False)
        self._setshifter(False)  # it is not a shifter type of component

    def calculate(self):
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        p2 = self.parameters[2]
        p3 = self.parameters[3]
        p4 = self.parameters[4]
        p5 = self.parameters[5]

        if p0.ischanged() or p1.ischanged() or p2.ischanged() \
                or p3.ischanged() or p4.ischanged() or p5.ischanged():
            A1 = p0.getvalue()
            r1 = p1.getvalue()
            A2 = p2.getvalue()
            r2 = p3.getvalue()
            A3 = p4.getvalue()
            r3 = p5.getvalue()
            self.data = self.fast_background3(A1, r1, A2, r2, A3, r3)
        self.setunchanged()  # put parameters to unchanged

    def fast_background3(self, A1, r1, A2, r2, A3, r3):
        E = self.energy_axis
        Estart = E[0]
        if Estart < 1:
            Estart = 1

        mask = (E > 0)  # only meaningfull when E>0 and r>0
        return mask * (A1 * (E / Estart) ** (-r1) + A2 * (E / Estart) ** (
            -r2) + A3 * (E / Estart) ** (-r3))

    def getgradient(self, parameter):
        """calculate the analytical partial derivative wrt parameter j
        returns true if succesful, gradient is stored in component.gradient
        """
        p0 = self.parameters[0]
        p1 = self.parameters[1]
        p2 = self.parameters[2]
        p3 = self.parameters[3]
        p4 = self.parameters[4]
        p5 = self.parameters[5]

        r1 = p1.getvalue()
        r2 = p3.getvalue()
        r3 = p5.getvalue()

        E = self.energy_axis
        Estart = E[0]
        if Estart < 1:
            Estart = 1
        mask = (E > 0)  # only meaningfull when E>0 and r>0
        if parameter == p0:
            self.gradient[0] = mask * (E) ** (-r1)
            return self.gradient[0]
        elif parameter == p2:
            self.gradient[2] = mask * (E) ** (-r2)
            return self.gradient[2]
        elif parameter == p4:
            self.gradient[2] = mask * (E) ** (-r3)
            return self.gradient[4]
        else:
            # throw Componenterr::bad_index()
            return False



from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np


class Linear(Component):
    """
    A Linear component.
    """

    def __init__(self, specshape, m, q):
        """
        Initialize the Linear object
        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model.

        m : float
            The slope of the linear function.

        q: float
            The constant value of the linear function.

        Returns
        -------
        """
        super().__init__(specshape)

        p1 = Parameter('m', m)
        p1.setlinear(True)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('q', q)
        p2.setlinear(True)
        p2.sethasgradient(True)
        self._addparameter(p2)

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        if p1.ischanged() or p2.ischanged():
            m = p1.getvalue()
            q = p2.getvalue()
            self.data = self.linear_function(m, q)
        self.setunchanged()  # put parameters to unchanged

    def linear_function(self, m, q):
        return m * self.energy_axis + q

    def getgradient(self, parameter):
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        if parameter == p1:
            self.gradient[0] = self.linear_function(1, 0)
            return self.gradient[0]
        if parameter == p2:
            self.gradient[1] = np.ones(self.size)
            return self.gradient[1]
        return None



import sys

import numpy as np
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.model import Model

from pyEELSMODEL.components.linear import Linear
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.fitters.minimizefitter import MinimizeFitter
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT

sys.path.append("..")  # Adds higher directory to python modules path.


def test_MLE_perform_fit():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Gaussian(specshape, 10000, 400, 50)
    mod = Model(specshape, components=[pol])
    mod.calculate()
    s = Spectrum(specshape, data=np.random.poisson(mod.data))

    # values = [1,110,1]
    values = [10000, 400, 50]

    for param, value in zip(pol.parameters, values):
        param.setvalue(value)

    fit = MinimizeFitter(s, mod)
    fit.perform_fit()

    for value, coeff in zip(values, fit.coeff):
        rel_er = np.abs((coeff - value) / value)
        assert rel_er < 1e-2


def test_LSQ_perform_fit():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Gaussian(specshape, 10000, 400, 50)
    mod = Model(specshape, components=[pol])
    mod.calculate()
    s = Spectrum(specshape, data=np.random.normal(mod.data, 1))

    values = [10000, 400, 50]
    for param, value in zip(pol.parameters, values):
        param.setvalue(value)

    fit = MinimizeFitter(s, mod, estimator='LSQ')
    fit.perform_fit()

    for value, coeff in zip(values, fit.coeff):
        rel_er = np.abs((coeff - value) / value)
        assert rel_er < 1e-2


def test_jacobian():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Gaussian(specshape, 10000, 400, 50)
    gaus = Gaussian(specshape, 100, 300, 50)
    gaus.calculate()
    ll = MscatterFFT(specshape, gaus)

    mod = Model(specshape, components=[pol])
    mod.calculate()
    s = Spectrum(specshape, data=np.random.poisson(mod.data))
    fit = MinimizeFitter(s, mod)
    assert fit.jacobian != '2-point'

    mod1 = Model(specshape, components=[pol, ll])
    mod1.calculate()
    fit = MinimizeFitter(s, mod1)
    assert fit.jacobian == '2-point'

    fit = MinimizeFitter(s, mod, estimator='LSQ')
    assert fit.jacobian == '2-point'


def test_CRLB():
    specshape = Spectrumshape(1, 100, 1024)
    pol = Linear(specshape, m=1, q=1)
    mod = Model(specshape, components=[pol])
    mod.calculate()
    sig = np.copy(mod.data)

    nsamples = 100
    coeff_matrix = np.zeros((2, nsamples))

    for i in range(nsamples):
        pol.parameters[0].setvalue(1)
        pol.parameters[1].setvalue(1)

        ndata = np.random.poisson(sig)
        s = Spectrum(specshape, data=ndata)
        fit = MinimizeFitter(s, mod)
        fit.usegradients = False
        fit.perform_fit()
        coeff_matrix[:, i] = fit.coeff

    pol.parameters[0].setvalue(1)
    pol.parameters[1].setvalue(1)
    fit.set_information_matrix()

    std_a = np.std(coeff_matrix[0])
    crlb_a = fit.CRLB(pol.parameters[0])
    std_b = np.std(coeff_matrix[1])
    crlb_b = fit.CRLB(pol.parameters[1])

    assert np.abs(std_a - crlb_a) / (std_a + crlb_a) < 0.1
    assert np.abs(std_b - crlb_b) / (std_b + crlb_b) < 0.1


def main():
    test_MLE_perform_fit()
    test_LSQ_perform_fit()
    test_CRLB()
    test_jacobian()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import logging
import time
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.core.spectrum import Spectrum
from tqdm import tqdm
import matplotlib.pyplot as plt
from scipy import stats

logger = logging.getLogger(__name__)


class Fitter:
    """
    Fitter class from which all actual fitters can be derived.

    many things are in common for all fitters like:
        -calculate goodness of fit (actual implementation can be overridden eg
        difference between weighted least square or lsq or poisson specific)
        -do iterations until certain goodness of fit is reached
        -calculate degrees of freedom
        -calculate crlb
        -
    """

    def __init__(self, spectrum, model):
        """
        Initialises a Fitter instance.

        Parameters
        ----------
        spectrum : Spectrum or Multispectrum
            The experimental data used which needs to be fitted
        model : Model
            The model used to fit the experimental data.

        Returns
        -------
        An instance of a Fitter.

        """
        self.spectrum = spectrum
        self.model = model
        self.fittertype = 'fitter base class'
        self.status = 'status info'
        self.progress = 0
        self.stop = False
        self.set_start_parameters()
        self.set_bounds()

        if self.model.hasconvolutor():
            # no analytical gradient when using a convolution
            self.usegradients = False
        else:
            self.usegradients = True

        self.information_matrix = None
        self.covariance_matrix = None
        self.coeff_matrix = None
        self.coeff = None
        self.error = None
        self.error_matrix = None
        self.LR = None
        self.LR_matrix = None
        self.getcurrentfit = None
        if isinstance(self.spectrum, MultiSpectrum):
            # set the current spectrum to the first one
            self.spectrum.setcurrentspectrum(spectrum.currentspectrumid)
            self.getcurrentfit = spectrum.currentspectrumid

        self.elapsed_time = None  # time it took to do the multifit

    @property
    def fittertype(self):
        return self._fittertype

    @fittertype.setter
    def fittertype(self, fittertype):
        self._fittertype = fittertype

    # def getprogress(self):
    #     # returns an int from 0 to 100 indicating the progress in a multifit
    #     return self.progress

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    @property
    def usegradients(self):
        return self._usegradients

    @usegradients.setter
    def usegradients(self, b):
        # If convolution is performed, the analytical gradients are not used
        if self.model.hasconvolutor():
            logger.warning(
                'cannot use analytical gradients since a '
                'convolutor is inside model')
            self._usegradients = False

        else:
            self._usegradients = b

        # todo better implementation, will not work since jacobian
        # is not defined in the fitter.

        try:
            self.jacobian = self._jacobian
        except Exception:
            logger.info(r'No jacobian defined for this fitter type')

    @property
    def information_matrix(self):
        return self._information_matrix

    @information_matrix.setter
    def information_matrix(self, information_matrix):
        self._information_matrix = information_matrix

    @property
    def covariance_matrix(self):
        return self._covariance_matrix

    @covariance_matrix.setter
    def covariance_matrix(self, covariance_matrix):
        self._covariance_matrix = covariance_matrix

    def getlabelist(self):
        label_list = []
        for param in self.model.getfreeparameters():
            comp = self.model.getcomponentbyparameter(param)
            label_list.append(comp.name + ': ' + param.name)
        return label_list

    def show_covariance_matrix(self):
        """
        Shows the covariance matrix and adds the labels of the different
        components.

        Returns
        -------
        fig: Figure
        """

        self.set_information_matrix()
        label_list = self.getlabelist()

        fig, ax = plt.subplots()
        ax.imshow(np.abs(self.covariance_matrix))
        ax.set_xticks(np.arange(len(label_list)))
        ax.set_yticks(np.arange(len(label_list)))
        ax.set_xticklabels(label_list, rotation=45, ha="right")
        ax.set_yticklabels(label_list, rotation=45)
        fig.set_tight_layout(True)
        return fig

    def show_variance(self):
        """
        Shows the variance and adds the labels of the different components.

        Returns
        -------
        fig: Figure
        """
        label_list = self.getlabelist()
        print(label_list)
        fig, ax = plt.subplots()
        ax.plot(self.covariance_matrix.diagonal())
        ax.set_xticks(np.arange(len(label_list)))
        ax.set_xticklabels(label_list, rotation=45, ha="right")
        fig.set_tight_layout(True)
        return fig

    def show_coefficients(self, index=(0, 0)):
        if isinstance(self.spectrum, MultiSpectrum):
            self.setcurrentfit(index=index)

        label_list = self.getlabelist()

        fig, ax = plt.subplots()
        x = np.arange(self.coeff.size)
        width = x[1] - x[0]
        ax.bar(x, self.coeff, width=width)
        ax.set_xticks(np.arange(len(label_list)))
        ax.set_xticklabels(label_list, rotation=45, ha="right")
        fig.set_tight_layout(True)

    def getdoresidual(self):
        return self.residual

    def setminstep(self, x):
        self.minstep = x

    def setmaxstep(self, x):
        self.maxstep = x

    def setfraction(self, x):
        self.fraction = x

    def setusegradients(self, b):
        self.usegradients = b

    def setdoresidual(self, b):
        self.residual = b

    def settolerance(self, t):
        self.tolerance = t

    def setnmax(self, n):
        self.nmax = n

    def gettolerance(self):
        return self.tolerance

    def createmodelinfo(self):

        return 0

    def CRLB(self, par, index=None):
        """
        Returns the CRLB bound of the parameter par. If the spectrum is a
        multispectrum, the results for the different spectra can be modified by
        inputting another index.

        Parameters
        ----------
        par: Parameter
            A parameter which is used in the model to fit the data
        index: tuple
            The index of the spectrum from which we want the CRLB. Only works
            when the given spectrum is a multispectrum
            (default: use the current index).

        Returns
        -------
        CRLB: float
            The CRLB for the given parameter and spectrum.

        """
        if isinstance(self.spectrum, MultiSpectrum):
            if index is not None:
                self.setcurrentfit(index=index)

        ii = self.get_param_index(par)

        if ii is None:
            print("This parameter is not fitted, check if this is not free or"
                  " non existing in model")
            return None

        if self.covariance_matrix is None:
            self.set_information_matrix()

        CRLB = np.sqrt(self.covariance_matrix[ii, ii])
        return CRLB

    def CRLB_map(self, par):
        """
        Returns the CRLB bound of the parameter par. For each scanned point.
        This only works when a multispectrum is fitted.

        Parameters
        ----------
        par: Parameter
            A parameter which is used in the model to fit the data

        Returns
        -------
        crlb_map: numpy array (2D)
            The CRLB map the given parameter.

        """

        if not isinstance(self.spectrum, MultiSpectrum):
            raise TypeError(r'Mapping not possible since given spectrum is no'
                            r' multispectrum')

        shape = (self.spectrum.xsize, self.spectrum.ysize)
        crlb_map = np.zeros(shape)
        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]
            ii = (islice[0], islice[1])
            crlb = self.CRLB(par, index=ii)
            crlb_map[islice] = crlb

        return crlb_map

    def CRLB_ratio(self, par1, par2, index=(0, 0)):
        """
        Returns the CRLB bound of the ratio between par1 and par2 (par1/par2).
        The results for the different spectra can be modified by inputting
        another index.

        Parameters
        ----------
        par1: Parameter
            A parameter which is used in the model to fit the data
        par2: Parameter
            A parameter which is used in the model to fit the data
        index: tuple
            The index of the spectrum from which we want the CRLB. Only works
            when the given spectrum is a multispectrum (default: (0,0)).

        Returns
        -------
        error: float
            The CRLB bound calculated for the ratio between two parameters.

        """

        if isinstance(self.spectrum, MultiSpectrum):
            self.setcurrentfit(self, index=index)

        ratio = par1.getvalue() / par2.getvalue()
        index1 = self.get_param_index(par1)
        index2 = self.get_param_index(par2)
        if index1 is None or index2 is None:
            print("This parameter is not fitted, check if this is not free or "
                  "non existing in model")
            return None

        if self.covariance_matrix is None:
            self.set_information_matrix()

        rel_er1 = self.covariance_matrix[index1, index1] / par1.getvalue() ** 2
        rel_er2 = self.covariance_matrix[index2, index2] / par2.getvalue() ** 2
        cov_er = 2 * self.covariance_matrix[index1, index2] / \
            (par1.getvalue() * par2.getvalue())
        error = ratio * np.sqrt(rel_er1 + rel_er2 - cov_er)
        return error

    def get_param_index(self, param):
        """
        Returns the index of the parameter in the free parameter list. This
        function is used to ask the fitted value for which parameter. If the
        given is not in the list it returns None

        Parameters
        ----------
        param: Parameter
            The parameter from which we want to get the index

        Returns
        -------
        index: int or None
            The index which corresponds to the given parameter in the fit.

        """
        for index, parameter in enumerate(self.model.getfreeparameters()):
            if parameter is param:
                return index
        return None

    def degreesoffreedom(self):
        """
        Returns the number of degrees of freedom in the fitter. This depends
        on the number of free parameters in the model and the number of fitted
        points in the spectrum
        """
        dof = self.spectrum.getnonexcludedpoints() \
            - self.model.getnumfreeparameters()
        return dof

    def LRtestconfidence(self, index=(0, 0)):
        """
        Confidence test using the likelyhood ratio of Poisson noise. See
        publication EELSmodel for explanation on this

        """
        if self.LR is None:
            logger.info('LR needs to be calculated')
            self.likelihood_ratio(index=index)

        dgf = self.degreesoffreedom()
        confidence = stats.chi2.cdf(self.LR, dgf)
        return confidence

    def multi_LRtestconficence(self):
        shape = (self.spectrum.xsize, self.spectrum.ysize)

        self.confidence_map = np.zeros(shape)

        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]
            self.confidence_map[islice] = self.LRtestconfidence(index)

    def LRtestconfidence_string(self):
        confidence = 100 * self.LRtestconfidence()
        print('The model is accepted at a confidence levet of: {conf} percent'
              .format(conf=confidence))

    def getstatus(self):
        return 'status string'

    def dolintrick(self, b):
        # wether to do a special lin/nonlin trick for the linear parameters
        self.dolintrick = b

    def getdolintrick(self):
        return self.dolintrick

    def candolintrick(self):
        return self.candolin

    def _setcandolin(self, b):
        self.candolin = b

    def set_information_matrix(self):
        """
        Calculates the fischer information matrix followed by formula (16)
        in https://doi.org/10.1016/j.ultramic.2004.06.004
        Used for Poisson noise

        """
        self.model.calculate()  # calculate the newest model

        # do not fully understand why **(-1) but it seems to fit
        ndata = self.spectrum.pppc ** (-1) * np.copy(self.model.data)
        # ndata[ndata < 1.] = 1. #small values will blow up lambda_m

        lambda_m = 1 / np.sqrt(np.tile(ndata[np.invert(self.spectrum.exclude)],
                                       (self.model.getnumfreeparameters(), 1)))
        deriv_matrix = lambda_m * self.calculate_derivmatrix()
        self.information_matrix = np.dot(deriv_matrix,
                                         np.transpose(deriv_matrix))

        try:
            self.covariance_matrix = np.linalg.inv(self.information_matrix)
        except Exception:
            # todo solve this issue since it can happen that two rows are zero
            print(
                'Information matrix is Singular, can happen if two components'
                ' are exactly zero')

    def calculate_derivmatrix(self):
        """
        Calculates the derivative matrix where each row is a derivative of the
        model spectrum to a free parameter. The number of columns depends on
        the size of the spectrum.

        Returns
        -------
        deriv_matrix: numpy array (2D)
            The derivative matrix for each fitted parameter.
        """
        deriv_matrix = np.zeros((self.model.getnumfreeparameters(),
                                 self.spectrum.getnonexcludedpoints()))
        for index, param in enumerate(self.model.getfreeparameters()):
            deriv_matrix[index] = self.partial_derivative(param)
        return deriv_matrix

    def partial_derivative(self, parameter, fraction=0.001):
        """
        Calculates the partial derivative of the parameter. It uses the
        use_gradient attribute to check if the gradient is calculated
        analytically or numerically.

        Parameters
        ----------
        parameter: Parameter
            The parameter from which the partial derivative will be
            calculated.
        fraction: float
            The fraction of the step the parameter does to calculate
            the numerical gradient. (default: 0.001)

        Returns
        -------
        parial: numpy array (1D)
            The partial derivative for the given parameter.
        """

        if not self.usegradients:
            partial = self.numerical_partial_derivative(parameter,
                                                        fraction=fraction)
        else:
            component = self.model.getcomponentbyparameter(parameter)
            partial = component.getgradient(parameter)
            if partial is None:
                logging.info(r'No analytical gradient is implemented for'
                             r' this parameter. It will use a numerical '
                             r'calculation to get the result')
                partial = self.numerical_partial_derivative(parameter,
                                                            fraction=fraction)

        if partial is None:
            logging.warning(r'No partial derivative can be determined')
            return None
        else:
            return partial[np.invert(self.spectrum.exclude)]

    def numerical_partial_derivative(self, parameter, fraction=0.001):
        """
        Calculates the numerical partial derivative of the parameter.

        Parameters
        ----------
        parameter: Parameter
            The parameter from which the numerical derivative will be
            calculated.
        fraction: float
            The fraction of the step the parameter does to calculate
            the numerical gradient (default: 0.001)

        Returns
        -------
        """

        minstep = 1e-99
        maxstep = 1e99
        component = self.model.getcomponentbyparameter(parameter)
        if component is None:
            return None
        if component._ismultiplier or component._isconvolutor \
                or component._isshifter:
            return None
        delt = np.abs(parameter.getvalue() * fraction)
        delt = max(delt, minstep)
        delt = min(delt, maxstep)

        orig_par = parameter.getvalue()
        self.model.calculate()
        orig_mod = self.model.data[:]

        parameter.setvalue(orig_par + delt)
        self.model.calculate()
        up_mod = self.model.data[:]
        parameter.setvalue(orig_par)
        partial = (up_mod - orig_mod) / delt
        return partial

    def set_start_parameters(self):
        """
        Gets the start parameters of the model as initial parameters for the
        fitting. This function is used when starting values need to be
        added to the fit.

        Returns
        -------
        start_param: List
            A list containing the starting values for the fitting

        """
        start_param = []
        for param in self.model.getfreeparameters():
            start_param.append(param.getvalue())
        self._start_param = start_param

    def get_start_parameters(self):
        return self._start_param

    def set_fit_values(self):
        """
        Function which set the fitting values to the model.
        """
        if self.coeff is None:
            logger.info('No fit has been performed hence the values cannot '
                        'be set.')
            return None

        for ii, param in enumerate(self.model.getfreeparameters()):
            # if setvalue does not work, then it returns False
            if not param.setvalue(self.coeff[ii]):
                logger.info(param.name + r' value cannot be set')
                param.setvalue(0)
        self.model.calculate()

    def multi_fit(self, start_param=None):
        """
        Performs the selected fit over each spectrum in the multispectrum.

        Parameters
        ----------
        start_param: list
            List of starting parameters for the fit. If None, the values
            of the model are used. The shape of the start_param should be
            of size (xsize, ysize, numberoffreeparamters). The starting
            parameters are only needed for the non-linear models.

        Returns
        -------

        """

        start = time.time()

        # iterate over the multispectrum
        if not isinstance(self.spectrum, MultiSpectrum):
            raise ValueError(r'Spectrum should be a multispectrum')

        # the convolutor used in the model
        conv = self.model.getconvolutor()

        shape = (self.spectrum.xsize, self.spectrum.ysize)
        self.progress = 0
        self.stop = False
        initialindex = self.spectrum.currentspectrumid

        step = 100 / (self.spectrum.xsize * self.spectrum.ysize)
        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]
            self.spectrum.setcurrentspectrum(index)

            # if the low loss spectrum is a multispectrum, the index needs to
            # be changed
            if conv is not None:
                if isinstance(conv.llspectrum, MultiSpectrum):
                    conv.llspectrum.setcurrentspectrum(index)

            # need to be sure that the order of the free parameters is the same
            # as the input start_param
            if start_param is not None:
                self._start_param = start_param[islice]

            # small functionality which should give the ability to use "nearby"
            # solution to have a better initial guess
            # not well tested since mainly linear fitters are used and this
            # does not need any starting parameters
            # if use_prev:
            #     self.set_start_parameters()

            if self.stop:
                # restore to state that we started from
                self.spectrum.setcurrentspectrum(initialindex)
                if conv is not None:
                    if isinstance(conv.llspectrum, MultiSpectrum):
                        conv.llspectrum.setcurrentspectrum(initialindex)
                return

            self.perform_fit()

            # initialized the result matrix at first iteration of the fit.
            if (islice[0] == 0) and (islice[1] == 0):
                self.coeff_matrix = np.zeros((shape[0], shape[1],
                                              self.coeff.size))
                self.error_matrix = np.zeros((shape[0], shape[1]))

            self.coeff_matrix[islice] = self.coeff
            self.error_matrix[islice] = self.error

            self.progress = self.progress + step

        stop = time.time()
        self.elapsed_time = stop - start  # calculates the calculation time

    def set_bounds(self):
        """
        Sets the boundaries for the fitting. It uses the boundaries from
        each free parameter in the model to get these values.
        """
        bounds = []
        for param in self.model.getfreeparameters():
            bounds.append((param.getlowerbound(), param.getupperbound()))
        self._bounds = bounds

    def get_bounds(self):
        return self._bounds

    def model_to_multispectrum(self):
        """
        After the fitting of the multispectrum with the given model, the
        resulting model at every probe position is calculated which will be a
        multispectrum.

        Returns
        -------
        sig: MultiSpectrum
            The fitted multispectrum at each probe position.
        """
        conv = self.model.getconvolutor()

        sig = self.spectrum.copy()
        m = self.model
        shape = (sig.xsize, sig.ysize)
        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]

            if conv is not None:
                if isinstance(conv.llspectrum, MultiSpectrum):
                    conv.llspectrum.setcurrentspectrum(index)

            for ii, param in enumerate(m.getfreeparameters()):
                param.setvalue(self.coeff_matrix[islice][ii])
            m.calculate()
            sig.multidata[islice] = m.data

        return sig

    def model_to_multispectrum_without_comps(self, comps):
        """
        Set the parameter values to zero for the components in comps and
        calculates the fitted model. This can be used to remove the background
        from the fitted result or to only have the background.

        Parameters
        ----------
        comps: list of Components
            List of components which should not be added to the calcualted
            model.

        Returns
        -------
        sig: MultiSpectrum
            The fitted multispectrum at each probe position without the given
            components in comp.
        """
        #suppress the components you don't want
        for comp in comps:
            comp.setsuppress(True)
        self.model.setchanged(True)
        sig=self.model_to_multispectrum()
        
        #release the suppress to not confuse things
        for comp in comps:
            comp.setsuppress(False)

        return sig
    
    def model_to_multispectrum_with_comps(self, comps):
        """
        Only uses the components in comps to calculate the fitted model.
        Similar to model_to_multispectrum_without_comps but the exact opposite
        Note that this only works 1st level components and not with subcomponents owned by these 1st level components

        Parameters
        ----------
        comps: list of Components
            List of components are only used in the calculation of the fit.

        Returns
        -------
        sig: MultiSpectrum
            The fitted multispectrum at each probe position only have the given
             components in comps.
        """
        ncomps = []
        for comp in self.model.components:
            if comp in comps:
                pass
            else:
                ncomps.append(comp)

        s = self.model_to_multispectrum_without_comps(ncomps)
        return s

    def get_experimental_edge(self, component, percentile, plotting=False,
                              other_spectra=[]):
        """
        Calculate the proposed edge when using a multispectrum. The percentile
        takes the spectra having the highest content of these edges.


        :return:
        """
        if not isinstance(self.spectrum, MultiSpectrum):
            raise ValueError(r'Inputted spectrum should be a multispectrum')

        if self.coeff_matrix is None:
            raise ValueError(r'Fit is not yet performed, this should be done '
                             r'before getting'
                             r'proposed experimental edge')

        sig_w_comp = self.model_to_multispectrum_without_comps([component])

        index = self.get_param_index(component.parameters[0])
        print(index)
        int_map = self.coeff_matrix[:, :, index]
        co = int(percentile * int_map.size)
        thres = np.sort(int_map.flatten())[co]
        boolean = int_map > thres

        if plotting:
            plt.figure()
            plt.imshow(boolean, cmap='gray')
            plt.title('The mask used to determine the average ')

        avg_ = self.spectrum.multidata[boolean, :]

        res = np.sum(avg_ - sig_w_comp.multidata[boolean, :], axis=0)
        spec = Spectrum(self.spectrum.get_spectrumshape(), data=res)
        avg = Spectrum(self.spectrum.get_spectrumshape(),
                       data=avg_.sum(axis=0))

        other = []
        for ospec in other_spectra:
            avg_ = ospec.multidata[boolean, :]
            other.append(
                Spectrum(ospec.get_spectrumshape(), data=avg_.sum(axis=0)))

        return spec, avg, other

    def plot(self, index=None, externalplt=None, non_components=[], **kwargs):
        """
        Plots the fitter

        Parameters
        ----------
        externalplt : matplotlib reference
              A reference to an external matplotlib reference, if None we use
              our own matplotlib and create a new figure.

        Returns
        ----------
        fig: Figure
        """
        tempplt = plt
        if isinstance(externalplt, plt.Figure):
            tempplt = externalplt
        elif isinstance(externalplt, plt.Axes):
            tempplt = externalplt
            print('axes as input')
        else:
            # create our own figure
            fig = plt.figure()
            plt.title('Fit')
        # show components if visible
        # self.model.calculate()
        if isinstance(self.spectrum, MultiSpectrum) and index is not None:
            self.setcurrentfit(index=index)

        conv = self.model.getconvolutor()
        if conv is not None:
            if isinstance(conv.llspectrum, MultiSpectrum):
                conv.llspectrum.setcurrentspectrum(index)

        self.set_fit_values()
        tempplt.plot(self.spectrum.energy_axis, self.spectrum.data,
                     color='black', label='Experiment', linestyle='dotted')
        for comp in self.model.components:
            if not (comp in non_components):
                tempplt.plot(self.model.energy_axis, comp.data,
                             label=comp.name,
                             **kwargs)

            # print(comp.name)
            # print(np.sum(comp.data))
            # print('- - - - - ')

        tempplt.plot(self.model.energy_axis, self.model.data, label='Model',
                     **kwargs)  # and the total spectrum

        tempplt.fill_between(self.spectrum.energy_axis, 0,
                             self.spectrum.data.max(),
                             where=self.spectrum.exclude, color='green',
                             alpha=0.5)

        if isinstance(externalplt, plt.Axes):
            tempplt.set_xlabel(r'Energy Loss [eV]')
            tempplt.set_ylabel('Counts')
        else:
            tempplt.xlabel(r'Energy Loss [eV]')
            tempplt.ylabel('Counts')
        tempplt.legend()
        if isinstance(self.spectrum, MultiSpectrum):
            tempplt.title(self.spectrum.currentspectrumid)

        if externalplt is None:
            return fig
        else:
            return None

    def likelihood_ratio(self, index=(0, 0)):
        """
        Defined likelihood ratio from: 10.1016/j.ultramic.2004.06.004
        The index will be used when using a multispectrum.

        Parameters
        ----------
        index : tuple
              A tuple showing the index of the used spectrum for the likelihood
              ratio calculation.

        Returns
        ----------
        None

        """
        if isinstance(self.spectrum, MultiSpectrum):
            self.setcurrentfit(index=index)

        data = self.spectrum.data[np.invert(self.spectrum.exclude)].copy()
        data[data <= 0] = 1
        model = self.model.data[np.invert(self.spectrum.exclude)]

        LR = 2 * np.sum(-data + data * np.log(data / model) + model)
        self.LR = LR

    def multi_likelihood_ratio(self):
        shape = (self.spectrum.xsize, self.spectrum.ysize)

        self.LR_map = np.zeros(shape)

        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]
            self.likelihood_ratio(index)
            self.LR_map[islice] = self.LR

    def setcurrentfit(self, index=(0, 0)):
        """
        Set the fit to the used index. It will recalculate the model and
        information matrix for the used index. If the index has not changed, it
        will not perform any calculation to reduce the calculation time.

        Parameters
        ----------
        index : tuple
              A tuple showing the index of the used spectrum.

        Returns
        ----------
        None

        """
        if not isinstance(self.spectrum, MultiSpectrum):
            print('Spectrum is no MultiSpectrum hence this method has no use')
            return None

        if index == self.getcurrentfit and self.information_matrix is not None:
            pass
        else:
            self.getcurrentfit = index
            self.spectrum.setcurrentspectrum(index)
            self.coeff = self.coeff_matrix[index[0], index[1]]
            self.set_fit_values()
            self.set_information_matrix()

    def show_map_result(self, comp_elements):
        """
        Shows the maps for the given components in the list. Note that it only
        uses the first free parameter of the component. Hence this is mainly
        used for visualization of the elements maps where the components of
        the coreloss edges have only one parameter which needs to be fitted.


        Parameters
        ----------
        comp_elements : list
              List containing the components used in the model which have been
              fitted.

        Returns
        ----------
        fig: matplotlib figure
            The figure which is created. This way you can save it or modify it
            if needed.
        maps: numpy array
            Contains all the different maps where the indices are the same as
            the list of components.
        names: list
            List containing the name for each map. These returns are mainly
            used for modifying the images if more advanced figures need to be
            created for publications or reports.

        """
        maps = np.zeros((len(comp_elements), self.spectrum.xsize,
                         self.spectrum.ysize))
        names = np.zeros(len(comp_elements), dtype='object')

        # todo ugly way of discriminating between 1 or multiple components
        if len(comp_elements) == 1:
            fig, ax = plt.subplots(1, len(comp_elements), figsize=(8, 3))
            for i in range(len(comp_elements)):
                index = self.get_param_index(comp_elements[i].parameters[0])
                ax.imshow(self.coeff_matrix[:, :, index], cmap='inferno')
                ax.set_title(comp_elements[i].name)
                ax.set_xticks([])
                ax.set_yticks([])
                maps[i] = self.coeff_matrix[:, :, index]
                names[i] = comp_elements[i].name

        else:
            fig, ax = plt.subplots(1, len(comp_elements), figsize=(8, 3))
            for i in range(len(comp_elements)):
                index = self.get_param_index(comp_elements[i].parameters[0])
                ax[i].imshow(self.coeff_matrix[:, :, index], cmap='inferno')
                ax[i].set_title(comp_elements[i].name)
                ax[i].set_xticks([])
                ax[i].set_yticks([])
                maps[i] = self.coeff_matrix[:, :, index]
                names[i] = comp_elements[i].name

        return fig, maps, names

    def get_map_results(self, comp_elements):
        """
        Calculates the maps of the elements.

        Parameters
        ----------
        comp_elements : list
              List containing the components used in the model which have been
              fitted.

        Returns
        ----------
        maps: numpy array
            Contains all the different maps where the indices are the same as
            the list of components.
        names: list
            List containing the name for each map. These returns are mainly
            used for modifying the images if more advanced figures need to be
            created for publications or reports.
        """
        maps = np.zeros((len(comp_elements), self.spectrum.xsize,
                         self.spectrum.ysize))
        names = np.zeros(len(comp_elements), dtype='object')
        for i in range(len(comp_elements)):
            index = self.get_param_index(comp_elements[i].parameters[0])
            maps[i] = self.coeff_matrix[:, :, index]
            names[i] = comp_elements[i].name
        return maps, names

    def get_jump_ratio_map(self, component, onset_energy, interval=10):
        """
        Returns a map with the jump ratio between the background and edge
        calculated by integrating over interval region.
        This is an experimental function which needs to be more tested and
        explored.
        """

        sig_bkg = self.model_to_multispectrum_without_comps([component])
        sig_comp = self.model_to_multispectrum_with_comps([component])

        index = sig_bkg.get_energy_index(onset_energy)

        bkg_int = sig_bkg.multidata[:, :, index:index + interval].sum(2)
        comp_int = sig_comp.multidata[:, :, index:index + interval].sum(2)

        jump_ratio = comp_int / bkg_int

        return jump_ratio, comp_int, bkg_int, sig_bkg, sig_comp

    def get_map(self, param):
        """
        Returns the fitted coefficients of the parameter param. This only
        work if a multispectrum is used to fit.

        Parameters
        ----------
        param : Parameter
              A parameter which is optimized and from which the resulting
              map is requested.

        Returns
        ----------
        param_map: 2d numpy array
            The coefficient map of the given parameters

        """
        index = self.get_param_index(param)
        if index is None:
            print('Parameter is not optimized during the fit')
            return None
        param_map = self.coeff_matrix[:, :, index]
        return param_map



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np


class Offset(Component):
    """
    An offset component
    """

    def __init__(self, specshape, A):
        """
        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            The value of the offset.

        Returns
        -------
        """
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(-np.inf, np.inf)
        p1.sethasgradient(True)
        self._addparameter(p1)
        self.name = 'Offset'

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        if p1.ischanged():
            A = p1.getvalue()
            self.data = A * np.ones(self.size)
        self.setunchanged()  # put parameters to unchanged

    def getgradient(self, parameter):
        p1 = self.parameters[0]
        if parameter == p1:
            self.gradient[0] = np.ones(self.energy_axis)
            return self.gradient[0]
        return None



import numpy as np
from scipy import interpolate

import pyEELSMODEL.misc.physical_constants as pc
import pyEELSMODEL.misc.hydrogen_gdos as hdos
import warnings


def getgosq(info1_1, info1_2, n):
    return info1_1 * (np.exp(np.arange(n) * info1_2) - 1) * 1e10


def getgosenergy(info2_1, info2_2, n):
    return info2_1 * (np.exp(np.arange(n) * info2_2 / info2_1) - 1)


def linear_interpolation(x, y, x1):
    m = (y[1] - y[0]) / (x[1] - x[0])
    q = y[0] - m * x[0]
    return m * x1 + q


def powerlaw(x, A, r):
    return A * x ** (-r)


def getinterpolatedgos(E, q, E_axis, q_axis, GOSmatrix, swap_axes=False):
    """
    Gets the interpolated value of the GOS from the E and q value.


    Parameters
    ----------
    E: float
        The energy from which the GOS should be interpolated
    q: float
        The q from the GOS should be interpolated
    E_axis: numpy array
        The energy axis on which the GOS is calculated
    q_axis: numpy array
        The q axis on which the GOS is calculated
    swap_axes: boolean
        Indicates if the axes for q and E should be swapped. There is a
        difference between the GOS from Segger and Zhang.

    Returns
    -------
    interpolated GOS matrix

    """
    index_q = np.searchsorted(q_axis, q, side='left')
    index_E = np.searchsorted(E_axis, E, side='left')

    if index_E == 0:
        return 0
    if index_E == E_axis.size:
        return 0.
    if index_q == 0:
        return GOSmatrix[index_E, 0]
    if index_q == q_axis.size:
        return 0.0

    dE = E_axis[index_E] - E_axis[index_E - 1]
    dq = q_axis[index_q] - q_axis[index_q - 1]

    distE = E - E_axis[index_E - 1]
    distq = q - q_axis[index_q - 1]

    if swap_axes:
        r0 = GOSmatrix[index_q - 1, index_E - 1] * (1 / (dE * dq)) * (
                    dE - distE) * (dq - distq)
        r1 = GOSmatrix[index_q, index_E - 1] * (1 / (dE * dq)) * (
                    dE - distE) * (distq)
        r2 = GOSmatrix[index_q - 1, index_E] * (1 / (dE * dq)) * (distE) * (
                    dq - distq)
        r3 = GOSmatrix[index_q, index_E] * (1 / (dE * dq)) * (distE) * (distq)
    else:
        r0 = GOSmatrix[index_E - 1, index_q - 1] * (1 / (dE * dq)) * (
                    dE - distE) * (dq - distq)
        r1 = GOSmatrix[index_E - 1, index_q] * (1 / (dE * dq)) * (
                    dE - distE) * (distq)
        r2 = GOSmatrix[index_E, index_q - 1] * (1 / (dE * dq)) * (distE) * (
                    dq - distq)
        r3 = GOSmatrix[index_E, index_q] * (1 / (dE * dq)) * (distE) * (distq)

    return r0 + r1 + r2 + r3


def get_powerLaw_extrapolation(rel_energy_axis, q_axis, GOSmatrix, E0, beta,
                               alpha,
                               q_steps=100, swap_axes=False):
    """
    If the energy axis for the cross section extends the range of calculated
    GOS then the last two point are used to extrapolate a power-law from.


    Parameters
    ----------
    rel_energy_axis: float
        The energy from which the GOS should be interpolated
    q: float
        The q from the GOS should be interpolated
    E_axis: numpy array
        The energy axis on which the GOS is calculated
    q_axis: numpy array
        The q axis on which the GOS is calculated
    swap_axes: boolean
        Indicates if the axes for q and E should be swapped. There is a
        difference between the GOS from Segger and Zhang.

    Returns
    -------
    A: float
        The amplitude of the power-law
    r: float
        The power of the power-law
    """
    E_axis = rel_energy_axis[-2:]
    dsigma_dE = np.zeros(E_axis.size)
    R = pc.R()
    T = pc.T(E0)
    gamma = pc.gamma(E0)
    for i in range(E_axis.size):
        E = E_axis[i]
        integral = 0

        qa0sq_min, qa0sq_max = hdos.get_qmin_max(E, E0, beta, alpha=alpha)
        logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                    q_steps)
        lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
        for j in range(logqa0sq_axis.size):
            q = np.sqrt(np.exp(logqa0sq_axis[j])) / pc.a0()
            theta = 2. * np.sqrt(np.abs(
                R * (np.exp(logqa0sq_axis[j]) - qa0sq_min) / (
                            4. * gamma ** 2 * T)))
            df_dE = getinterpolatedgos(E, q, rel_energy_axis, q_axis,
                                       GOSmatrix, swap_axes=swap_axes)
            # integral += df_dE * lnqa0sqstep
            integral += df_dE * lnqa0sqstep * hdos.correction_factor_kohl(
                alpha, beta, theta)
        dsigma_dE[i] = 4 * np.pi * pc.a0() ** 2 * (R / E) * (R / T) * integral

    r = np.log(dsigma_dE[1] / dsigma_dE[0]) / np.log(E_axis[0] / E_axis[1])
    A = dsigma_dE[0] / (pow(E_axis[0], -r))

    return A, r


# def test_getinterpolatedgos():
#     z_matrix = np.ones((50,50))
#     x = np.arange(z_matrix.shape[0])
#     y = np.arange(z_matrix.shape[1])
#
#     z_matrix[11,9] = 2
#     z_matrix[12,9] = 4
#     z_matrix[11,10] = 2
#     z_matrix[12,10] = 4
#
#     print(getinterpolatedgos(11.5, 9.5, x, y, z_matrix))


def dsigma_dE_HS(energy_axis, Z, ek, E0, beta, alpha, filename, q_steps=100):
    """
    Calculates the cross section from the Hartree-Slater cross sections
    from Rez. They are not used since they are not open-access.

    """

    with open(filename) as f:
        GOS_list = f.read().replace('\r', '').split()
    R = pc.R()
    # Map the parameters
    q_1 = float(GOS_list[2])
    q_2 = float(GOS_list[3])
    ncol = int(GOS_list[5])
    E_1 = float(GOS_list[6])
    E_2 = float(GOS_list[7])
    nrow = int(GOS_list[8])
    gos_array = np.array(GOS_list[9:], dtype=np.float64)
    # careful data is stored per rydberg and we want
    # per eV to be compatible with sigmak
    GOSmatrix = gos_array.reshape(nrow, ncol) / R

    rel_energy_axis = getgosenergy(E_1, E_2, nrow) + ek
    q_axis = getgosq(q_1, q_2, ncol)
    dsigma_dE = dsigma_dE_from_GOSarray(energy_axis, rel_energy_axis, ek, E0,
                                        beta, alpha, q_axis,
                                        GOSmatrix, q_steps=q_steps)
    return dsigma_dE


def dsigma_dE_from_GOSarray(energy_axis, rel_energy_axis, ek, E0, beta, alpha,
                            q_axis, GOSmatrix, q_steps=100, swap_axes=False):
    """
    Calculates the cross section from the GOS array. The integral over q-axis
    is done on a logarithmic scale.
    Note the speed of the calculation is not super-fast and could be improved
    by a better implementation of the interpolation and integration.

    Parameters
    ----------
    energy_axis: 1d numpy array
        The energy axis on which the cross section is calculated. [eV]
    rel_energy_axis: 1d numpy array
        The energy axis on which the GOS table is calculated. [eV]
    ek: float
        The onset energy of the calculated edge [eV]
    E0: float
        The acceleration voltage of the incoming electrons [V]
    alpha: float
        The convergence angle of the incoming probe [rad]
    beta:
        The collection angle of the outgoing electrons [rad]
    q_axis: 1d numpy array
        The momentum on which the GOS table are calculated. [kg m /s]?
    GOSmatrix: 2d numpy array
        The GOS
    q_steps: uint
        The number of q points used to numerically calculate the integral
        over the q direction. (default: 100)
    swap_axes: boolean
        The two GOS tables from Segger and Zhang have different axes. So
        for one, the energy axis is the first one and for the other it is
        the q-axis. Hence by swapping the axes we can use the same function
        for both. (default: False)

    Returns
    -------
    dsigma_dE: 1d numpy array
        The calculated cross section in m^2

    """
    R = pc.R()
    T = pc.T(E0)
    gamma = pc.gamma(E0)

    dsigma_dE = np.zeros(energy_axis.size)

    # check if there are energies larger then the max energy
    if energy_axis[-1] > rel_energy_axis[-1]:
        # then calculate a power law dependence of the cross sections
        powA, powr = get_powerLaw_extrapolation(rel_energy_axis, q_axis,
                                                GOSmatrix,
                                                E0, beta, alpha,
                                                q_steps=q_steps,
                                                swap_axes=swap_axes)

    for i in range(energy_axis.size):
        E = energy_axis[i]
        integral = 0
        if (E > ek) & (E <= rel_energy_axis[-1]):
            qa0sq_min, qa0sq_max = hdos.get_qmin_max(E, E0, beta, alpha=alpha)
            logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                        q_steps)
            lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
            for j in range(logqa0sq_axis.size):
                q = np.sqrt(np.exp(logqa0sq_axis[j])) / pc.a0()
                theta = 2. * np.sqrt(np.abs(
                    R * (np.exp(logqa0sq_axis[j]) - qa0sq_min) / (
                                4. * gamma ** 2 * T)))
                df_dE = getinterpolatedgos(E, q, rel_energy_axis, q_axis,
                                           GOSmatrix, swap_axes=swap_axes)
                # integral+= df_dE*lnqa0sqstep
                integral += df_dE * lnqa0sqstep * hdos.correction_factor_kohl(
                    alpha, beta, theta)
            # dsigma_dE[i] = 4*np.pi*pc.a0()**2*(R/E)*(R/T)*integral*dispersion
            dsigma_dE[i] = 4 * np.pi * pc.a0() ** 2 * (R / E) * (
                        R / T) * integral

        elif E > rel_energy_axis[-1]:
            dsigma_dE[i] = powerlaw(E, powA, powr)
        else:
            dsigma_dE[i] = 0

    return dsigma_dE


def dsigma_dE_from_GOSarray_approx(energy_axis, rel_energy_axis, E0, beta,
                                   GOSmatrix, swap_axes=False):
    """
    Calculates the cross section by using the approximation shown in
    https://doi.org/10.1016/0304-3991(89)90197-6, see Eq. 3
    """
    R = pc.R()
    T = pc.T(E0)
    # gamma = pc.gamma(E0)

    a0 = pc.a0()
    theta_E = pc.characteristic_angle(energy_axis, E0)

    # fac1 = 8 * a0**2 * R**2 / (energy_axis*pc.m0()*pc.speed_electron(E0)**2)
    fac1_ = 4 * np.pi * a0 ** 2 * (R / energy_axis) * (R / T)

    fac2 = np.log(1 + beta ** 2 / (theta_E ** 2))

    f = interpolate.interp1d(rel_energy_axis, GOSmatrix[:, 0], kind='linear',
                             fill_value='extrapolate')
    GOS = f(energy_axis)

    dsigma_dE = fac1_ * fac2 * GOS
    boolean = energy_axis < rel_energy_axis[0]
    dsigma_dE[boolean] = 0

    return dsigma_dE


def dsigma_dE_from_GOSarray_bound(energy_axis, free_energies, ek, E0, beta,
                                  alpha, q_axis, GOSmatrix, q_steps=100):
    """
    Calculates the cross section from the GOS array. The integral over q-axis
    is done on a logarithmic scale.
    This is a new method which is able to take the bounded states into account.
    Provided by Zezhong Zhang.

    Parameters
    ----------
    energy_axis: 1d numpy array
        The energy axis on which the cross section is calculated. [eV]
    free_energies: 1d numpy array
        The energy axis on which the GOS table is calculated without the onset
        energy [eV]
    ek: float
        The onset energy of the calculated edge [eV]
    E0: float
        The acceleration voltage of the incoming electrons [V]
    alpha: float
        The convergence angle of the incoming probe [rad]
    beta:
        The collection angle of the outgoing electrons [rad]
    q_axis: 1d numpy array
        The momentum on which the GOS table are calculated. [kg m /s]?
    GOSmatrix: 2d numpy array
        The GOS
    q_steps: uint
        The number of q points used to numerically calculate the integral
        over the q direction. (default: 100)
    swap_axes: boolean
        The two GOS tables from Segger and Zhang have different axes. So
        for one, the energy axis is the first one and for the other it is
        the q-axis. Hence by swapping the axes we can use the same function
        for both. (default: False)

    Returns
    -------
    dsigma_dE: 1d numpy array
        The calculated cross section in m^2

    """
    R = pc.R()
    T = pc.T(E0)
    gamma = pc.gamma(E0)

    # Define new energy axis which is uses the energies of the discreet
    # bounded states and in the continuum the usual energy axis.

    bool0 = free_energies < 0
    Ebound = free_energies[bool0] + ek

    dsigma_dE = np.zeros(energy_axis.size)
    dsigma_dE_bound = np.zeros(energy_axis.size)
    sigma = 2*(energy_axis[1] - energy_axis[0])

    rel_energy_axis = free_energies + ek
    # check if there are energies larger then the max energy
    if energy_axis[-1] > rel_energy_axis[-1]:
        # then calculate a power law dependence of the cross sections
        powA, powr = get_powerLaw_extrapolation(rel_energy_axis, q_axis,
                                                GOSmatrix,
                                                E0, beta, alpha,
                                                q_steps=q_steps,
                                                swap_axes=True)
    # the for loop over the bound states
    for i in range(Ebound.size):
        E = Ebound[i]

        integral = 0
        # the bounded states are differently interpolated
        qa0sq_min, qa0sq_max = hdos.get_qmin_max(E, E0, beta, alpha=alpha)
        logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                    q_steps)
        lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
        for j in range(logqa0sq_axis.size):
            q = np.sqrt(np.exp(logqa0sq_axis[j])) / pc.a0()
            theta = 2. * np.sqrt(np.abs(
                R * (np.exp(logqa0sq_axis[j]) - qa0sq_min) / (
                            4. * gamma ** 2 * T)))
            GOSarray = GOSmatrix[i, :]
            df_dE = getinterpolatedq(q, GOSarray, q_axis)

            # integral+= df_dE*lnqa0sqstep
            integral += df_dE * lnqa0sqstep * hdos.correction_factor_kohl(
                alpha, beta, theta)

        sig = 4 * np.pi * pc.a0() ** 2 * (R / E) * (R / T) * integral
        dsigma_dE_bound += gaussian(energy_axis, sig, E, sigma)

    # the for loop over the bound states
    for i in range(energy_axis.size):
        E = energy_axis[i]
        integral = 0
        if (E > ek) & (E <= rel_energy_axis[-1]):
            qa0sq_min, qa0sq_max = hdos.get_qmin_max(E, E0, beta, alpha=alpha)
            logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                        q_steps)
            lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
            for j in range(logqa0sq_axis.size):
                q = np.sqrt(np.exp(logqa0sq_axis[j])) / pc.a0()
                theta = 2. * np.sqrt(np.abs(
                    R * (np.exp(logqa0sq_axis[j]) - qa0sq_min) / (
                            4. * gamma ** 2 * T)))
                df_dE = getinterpolatedgos(E, q, rel_energy_axis, q_axis,
                                           GOSmatrix)
                # integral+= df_dE*lnqa0sqstep
                integral += df_dE * lnqa0sqstep * hdos.correction_factor_kohl(
                    alpha, beta, theta)
            # dsigma_dE[i] = 4*np.pi*pc.a0()**2*(R/E)*(R/T)*integral*dispersion
            dsigma_dE[i] = 4 * np.pi * pc.a0() ** 2 * (R / E) * (
                    R / T) * integral

        elif E > rel_energy_axis[-1]:
            dsigma_dE[i] = powerlaw(E, powA, powr)
        else:
            dsigma_dE[i] = 0

    return dsigma_dE + dsigma_dE_bound


def gaussian(E, integral, x0, sigma):
    # A = integral / (np.sqrt(2 * np.pi) * sigma)
    g = np.exp(-0.5 * (E - x0) ** 2 / sigma ** 2)
    g = integral * g / g.sum()
    return g


def getinterpolatedq(q, GOSarray, q_axis):
    """
    Gets the interpolated value of the GOS array as a function of q.
    Usefull for the bounded states


    Parameters
    ----------
    q: float
        The q from the GOS should be interpolated
    GOSarray:
        dddd
    q_axis: numpy array
        The q axis on which the GOS is calculated


    Returns
    -------
    interpolated GOS matrix

    """
    index_q = np.searchsorted(q_axis, q, side='left')

    if index_q == 0:
        return GOSarray[0]
    if index_q == q_axis.size:
        return 0.0

    dq = q_axis[index_q] - q_axis[index_q - 1]

    distq = q - q_axis[index_q - 1]

    r0 = GOSarray[index_q - 1] * (1/dq) * (dq-distq)
    r1 = GOSarray[index_q] * (1/dq) * (distq)

    return r0 + r1


def dsigma_dE_from_GOSarray_FastKohl(energy_axis, rel_energy_axis_, ek, E0, beta, alpha,
    q_axis, GOSmatrix_, q_steps=100):
    """
    Calculates the cross section from the GOS array. The integral over q-axis
    is done on a logarithmic scale.
    Note the speed of the calculation is imporved by a proper
    implementation of the interpolation and integration.

    Parameters
    ----------
    energy_axis: 1d numpy array
        The energy axis on which the cross section is calculated. [eV]
    rel_energy_axis: 1d numpy array
        The energy axis on which the GOS table is calculated. [eV]
    ek: float
        The onset energy of the calculated edge [eV]
    E0: float
        The acceleration voltage of the incoming electrons [V]
    alpha: float
        The convergence angle of the incoming probe [rad]
    beta:
        The collection angle of the outgoing electrons [rad]
    q_axis: 1d numpy array
        The momentum on which the GOS table are calculated. [kg m /s]?
    GOSmatrix: 2d numpy array
        The GOS
    q_steps: uint
        The number of q points used to numerically calculate the integral
        over the q direction. (default: 100)

    Returns
    -------
    dsigma_dE: 1d numpy array
        The calculated cross section in m^2

    """

    #add 0 for proper interpolation using the scipy method
    rel_energy_axis = np.pad(rel_energy_axis_-ek,(1,0))+ek #add 0 for proper interpolation using scipy method
    GOSmatrix = np.pad(GOSmatrix_,((0,0),(1,0),(0,0)))
    R = pc.R()
    T = pc.T(E0)
    gamma = pc.gamma(E0)

    dsigma_dE = np.zeros(energy_axis.size)

    # check if there are energies larger then the max energy
    if energy_axis[-1] > rel_energy_axis[-1]:
        # then calculate a power law dependence of the cross sections
        powA, powr = get_powerLaw_extrapolation(rel_energy_axis, q_axis,
                                                GOSmatrix,
                                                E0, beta, alpha,
                                                q_steps=q_steps,
                                                swap_axes=True)

    regime1 = np.logical_and(energy_axis > ek, energy_axis <= rel_energy_axis[-1])
    regime2 = energy_axis > rel_energy_axis[-1]
    regime3 = energy_axis <= ek


    dsigma_dE [regime1] = _FastKohl(energy_axis[regime1], E0, beta, alpha,q_steps,pc,T,gamma,rel_energy_axis,q_axis,GOSmatrix,R)

    if regime2.any():
            dsigma_dE[regime2]=powerlaw(energy_axis[regime2], powA, powr)

    dsigma_dE[regime3]=0

    return dsigma_dE

def _FastKohl(E, E0, beta, alpha,q_steps,pc,T,gamma,rel_energy_axis,q_axis,GOSmatrix,R):

    qa0sq_min, qa0sq_max = hdos.get_qmin_max(E, E0, beta, alpha=alpha)

    logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                q_steps)
    lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
    qs = np.sqrt(np.exp(logqa0sq_axis)) / pc.a0()
    
    thetas = 2. * np.sqrt(np.abs(R * (np.exp(logqa0sq_axis) - qa0sq_min) / (4. * gamma ** 2 * T)))

    rgi = interpolate.RegularGridInterpolator((q_axis,rel_energy_axis),GOSmatrix[:,:,0],)
    
    es = np.repeat(E[np.newaxis,:],q_steps,0)
    points = np.dstack((qs,es)).reshape((-1,2))
    out = rgi(points)


    integral=out.reshape(qs.shape)*_Fast_correction_factor_kohl(alpha, beta, thetas)

    integral=integral.sum(0)
    integral*=lnqa0sqstep


    integral[E<rel_energy_axis[1]]=0# gos and e_axis are padded for a proper behaviour of RegularGridInterpolator. This sets to 0 values affected by this.

    return 4 * np.pi * pc.a0() ** 2 * (R / E) * (R / T) * integral

def _Fast_correction_factor_kohl(alpha, beta, theta, min_alpha=1e-6):
    """
    STILL NEEDS TO BE VALIDATED
    Calculates the correction factor when using a convergent
    probe. For probes having is convergence angle smaller than
    min_alpha no correction is applied.
    Ultramicroscopy 16 (1985) 265-268:
    https://doi.org/10.1016/0304-3991(85)90081-6
     Parameters
     ----------
     alpha : float
         Convergence angle in radians
     beta : float
         Collection angle in radians
     theta : float
         The angle for which the correction factor should be calculated
     min_alpha : float
        Minimum convergence angle for which the correction is applied

     Returns
     -------
     corr_factor : float
        correction factor used in the integration
    """
    thetasq = theta ** 2
    alphasq = alpha ** 2
    betasq = beta ** 2

    
    if alpha < min_alpha:
        corr_factor = 1.
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")# invalid values for the correction are calculated, but tossed by np.where.
            corr_factor = np.where(theta <= np.abs(alpha - beta),min(alphasq, betasq)/ alphasq, _Fast_Fcorr_factor(alpha,beta,theta,thetasq,alphasq,betasq))
    return corr_factor
    
def _Fast_Fcorr_factor(alpha,beta,theta,thetasq,alphasq,betasq):
    x = (alphasq + thetasq - betasq) / (2. * alpha * theta)
    y = (betasq + thetasq - alphasq) / (2. * beta * theta)
    wortel = np.sqrt(4 * alphasq * betasq - (alphasq + betasq - thetasq) ** 2)
    corr_factor = (1 / np.pi) * (np.arccos(x) + (betasq / alphasq * np.arccos(y)) - (1 / (2 * alphasq) * wortel))
    return corr_factor





# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import matplotlib.pyplot as plt
import logging
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.model import Model
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.core.spectrum import Spectrum

from pyEELSMODEL.components.powerlaw import PowerLaw
from pyEELSMODEL.components.fast_background import FastBG2
from pyEELSMODEL.components.polynomial import Polynomial
from pyEELSMODEL.components.exponential import Exponential
from pyEELSMODEL.components.CLedge.zezhong_coreloss_edgecombined import\
    ZezhongCoreLossEdgeCombined
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT
from pyEELSMODEL.fitters.lsqfitter import LSQFitter
from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.fitters.minimizefitter import MinimizeFitter
from tqdm import tqdm

logger = logging.getLogger(__name__)


class BackgroundRemoval(Operator):
    """
    BackgroundRemoval is a class which needs a spectrum and signal range.
    This class provides a workflow on going extracting the elemental map
    of one edge.
    """

    def __init__(self, spectrum, signal_range, model_type='Powerlaw',
                 linear_fitting=False, order=2,
                 non_linear_fitter='LSQ', r_values=(2, 3)):
        """
        Initiates the BackgroundRemoval object

        Parameters
        ----------
        spectrum: Spectrum
            The spectrum from which the background should be removed
        signal_range: tuple
            Indicates the region on which the background fitting should be
            performed. This object can perform the fitting on two area and
            then two tuples should be given.
        model_type: str
            Indicates which model is used to perform the background fitting.
            At this point Powerlaw, FastBG, Exponential en Polynomial is
            implemented. (default: Powerlaw)
        linear_fitting: bool
            Indicates if a linear fitting procedure is used to determine the
            background. This can only work if the model only includes linear
            parameters such as for FastBG or Polynomial.
            (default: False)
        order: int
            If the model_type is 'Polynomial', the order value indicates the
            order of the polynomial. (default:2)
        non_linear_fitter: 'ML' or 'LSQ'
            The type of nonlinear fitter used in the fit. (default: LSQ)
        r_values: tuple of ints
            When the background model is the FastBG, the r values can be
            chosen. (default: (2,3))

        """

        self.spectrum = spectrum
        self.signal_range = signal_range
        self.model_type = model_type
        self.do_linear_fitting = linear_fitting
        self.non_linear_fitter = non_linear_fitter
        self.order = order
        self._r_values = r_values

        self.set_indices()
        self.make_background_model()
        self.set_fit_type()
        self.progress = 0
        self.result = None
        self._fast_fit = None
        self._multi_fast_fit = None
        self.multi_model_signal = None
        self.result = None

    @property
    def signal_range(self):
        return self._signal_range

    @signal_range.setter
    def signal_range(self, signal_range):
        """
        Set the signal range and first validates if the given signal_range is
        valid. The signal range can have two regions such that a interpolation
        on the background can be performed.

        Parameters
        ----------
        signal_range : tuple
            The signal range over which to fit the background.
        """
        # check if the signal range contains one or two ranges
        if any(isinstance(i, tuple) for i in signal_range):
            self.two_area = True
        else:
            self.two_area = False

        if not self.two_area:
            if signal_range[0] < self.spectrum.energy_axis[0] or \
                    signal_range[1] > self.spectrum.energy_axis[-1]:
                raise ValueError(r'Invalid signal range')
            if signal_range[0] >= signal_range[1]:
                raise ValueError(r'First number of signal range'
                                 r' must be smaller than the second value')

        else:
            for tup in signal_range:
                if tup[0] < self.spectrum.energy_axis[0] or \
                        tup[1] > self.spectrum.energy_axis[-1]:
                    raise ValueError(r'Invalid signal range')
                if tup[0] >= tup[1]:
                    raise ValueError(r'First number of signal range must'
                                     r' be smaller than the second value')

            if signal_range[0][1] > signal_range[1][0]:
                raise ValueError(r'The end energy of the first energy range '
                                 r'is larger than the beginning of the second')

        self._signal_range = signal_range

    @property
    def model_type(self):
        return self._model_type

    def getprogress(self):
        # returns an int from 0 to 100 indicating the progress in a multifit
        return self.fitter.progress

    def stop(self):
        # stop the fitting process
        self.fitter.stop = True

    @model_type.setter
    def model_type(self, model_type):
        model_list = ['Powerlaw', 'FastBG', 'Polynomial', 'Exponential']
        if model_type not in model_list:
            raise ValueError(r'Invalid model type')

        self._model_type = model_type

    @property
    def do_linear_fitting(self):
        return self._do_linear_fitting

    @do_linear_fitting.setter
    def do_linear_fitting(self, bool):
        if bool and (self.model_type != 'FastBG' and
                     self.model_type != 'Polynomial'):
            raise ValueError(r'This model type cannot perform a linear fit '
                             r'over the background. Set linear fitting to '
                             r'false if you want to use this type of'
                             r' background model')

        self._do_linear_fitting = bool

    @property
    def non_linear_fitter(self):
        return self._non_linear_fitter

    @non_linear_fitter.setter
    def non_linear_fitter(self, non_linear_fitter):
        non_list = ['LSQ', 'ML']
        if non_linear_fitter not in non_list:
            raise ValueError(r'Given non linear fitter is not implemented')

        self._non_linear_fitter = non_linear_fitter

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, order):
        if type(order) is not int:
            raise TypeError(r'Order should be integer')
        if order < 0:
            raise ValueError(r'Order should be a positive integer')

        self._order = order

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, m0):
        self._model = m0

    @property
    def fitter(self):
        return self._fitter

    @fitter.setter
    def fitter(self, fit):
        self._fitter = fit

    def set_indices(self):
        """
        Calculates the indices used which are excluded in the fit.
        These indices are also used to determine a first guess of the
        background model.
        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        if self.two_area:
            ind1 = [self.spectrum.get_energy_index(self.signal_range[0][0]),
                    self.spectrum.get_energy_index(self.signal_range[0][1])]
            ind2 = [self.spectrum.get_energy_index(self.signal_range[1][0]),
                    self.spectrum.get_energy_index(self.signal_range[1][1])]
            self.indices = [ind1, ind2]
        else:
            ind1 = [self.spectrum.get_energy_index(self.signal_range[0]),
                    self.spectrum.get_energy_index(self.signal_range[1])]
            self.indices = ind1

    def make_background_model(self):
        """
        Creates a model for the background, this depends on which model_type
        is chosen when creating the background object.
        The model will be set and given as an attribute to this object.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        specshape = self.spectrum.get_spectrumshape()
        m0 = Model(specshape)
        if self.model_type == 'Powerlaw':
            comp = PowerLaw(specshape, A=1, r=3)
        elif self.model_type == 'FastBG':
            comp = FastBG2(specshape, A1=1, r1=self._r_values[0], A2=1,
                           r2=self._r_values[1])
        elif self.model_type == 'Polynomial':
            comp = Polynomial(specshape, order=self.order)
        elif self.model_type == 'Exponential':
            comp = Exponential(specshape, A=1, b=-0.001)
        m0.addcomponent(comp)
        self.model = m0

    @property
    def fast_fit(self):
        return self._fast_fit

    @fast_fit.setter
    def fast_fit(self, param):
        self._fast_fit = param

    @property
    def multi_fast_fit(self):
        return self._multi_fast_fit

    @multi_fast_fit.setter
    def multi_fast_fit(self, param):
        self._multi_fast_fit = param

    def determine_fast_fit_parameters(self):
        """
        Performs the autofit over one spectrum, if the parameters need to be
        determined for a Multispectrum then the
        determine_multi_fast_fit_parameters() should be used.
        This is used as starting parameter for the more advanced fitting.
        Fast fitting only works for the powerlaw and exponential background
        models
        """
        if self.two_area:
            indic = [self.indices[0][0], self.indices[0][1]]
        else:
            indic = [self.indices[0], self.indices[1]]

        if self.model_type == 'Powerlaw' or self.model_type == 'Exponential':
            self.model.getcomponents()[0].autofit(self.spectrum,
                                                  indic[0],
                                                  indic[1])

            self.fast_fit = ([self.model.getfreeparameters()[0].getvalue(),
                              self.model.getfreeparameters()[1].getvalue()])

    def determine_multi_fast_fit_parameters(self):
        """
        Performs the autofit over the entire multispectrum.
        This is used as starting parameter for the more advanced fitting.
        Fast fitting only works for the powerlaw and exponential background
        models
        """
        print(isinstance(self.spectrum, MultiSpectrum))
        if not isinstance(self.spectrum, MultiSpectrum):
            raise ValueError(r'Spectrum used is no MultiSpectrum')

        if self.two_area:
            indic = [self.indices[0][0], self.indices[0][1]]
        else:
            indic = [self.indices[0], self.indices[1]]

        if self.model_type == 'Powerlaw' or self.model_type == 'Exponential':
            fast_matrix = np.empty((self.spectrum.xsize,
                                    self.spectrum.ysize, 2))
            for index in tqdm(np.ndindex(fast_matrix.shape[:-1])):
                islice = np.s_[index]
                self.spectrum.setcurrentspectrum(index)
                self.model.getcomponents()[0].autofit(self.spectrum,
                                                      indic[0], indic[1])
                A = self.model.getfreeparameters()[0].getvalue()
                r = self.model.getfreeparameters()[1].getvalue()
                fast_matrix[islice + (0,)] = A
                fast_matrix[islice + (1,)] = r

            self.multi_fast_fit = fast_matrix

        else:
            print('This model type does not support a fast fit method. '
                  'Only works for Powerlaw or Exponential')

    def include_areas(self):
        """
        Sets the exlude of the spectrum such that only the integration
        range is taken into account. It depends on the type of integration,
        it can be two area or just one area before the edge onset.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.spectrum.exclude = np.ones(self.spectrum.size, dtype=bool)
        if self.two_area:
            self.spectrum.set_include_region(self.indices[0][0],
                                             self.indices[0][1])
            self.spectrum.set_include_region(self.indices[1][0],
                                             self.indices[1][1])
        else:
            self.spectrum.set_include_region(self.indices[0], self.indices[1])

    def set_fit_type(self):
        """
        Set the fitter type for the background subtraction
        """
        if self.do_linear_fitting:
            fit = LinearFitter(self.spectrum, self.model)
        else:
            if self.non_linear_fitter == 'ML':
                fit = MinimizeFitter(self.spectrum, self.model)
                fit.method = 'SLSQP'
                # fit.estimator = fit.lsq_function

            elif self.non_linear_fitter == 'LSQ':
                fit = LSQFitter(self.spectrum, self.model,
                                use_bounds=True, method='trf')
        self.fitter = fit

    def calculate(self):
        """
        Returns the background subtracted spectrum. The exclude of the
        spectrum will be set the same as it was before this function was
        called. A least squares approach is used to fit the background.

        Returns
        -------
        self.results: Spectrum
            The background subtracted spectrum

        """
        prev_exclude = self.spectrum.exclude[:]
        self.set_indices()
        self.include_areas()
        # if self.fast_fit is None:
        #     logger.info('Perform the fast fitting to estimate parameters')
        self.determine_fast_fit_parameters()
        self.fitter.perform_fit()
        self.fitter.set_fit_values()
        self.spectrum.exclude = prev_exclude  # make the exclude is reset

        # if return_raw:
        #     return self.model.data
        # else:
        data = self.spectrum.data - self.model.data
        sback = Spectrum(self.spectrum.get_spectrumshape(), data=data)
        self.result = sback  # maybe get and setter?
        return self.result

    def calculate_multi(self):
        """
        Returns the background subtracted multispectrum. The exclude of the
        spectrum will be set the same as it was before this function was
        called.

        Returns
        -------
        self.results: MultiSpectrum
            The background subtracted multispectrum

        """

        prev_exclude = self.spectrum.exclude[:]
        self.set_indices()
        self.include_areas()

        if self.multi_fast_fit is None:
            logger.info('Perform the fast fitting to estimate parameters')
            self.determine_multi_fast_fit_parameters()

        self.fitter.multi_fit(start_param=self.multi_fast_fit)
        self.multi_model_signal = self.fitter.model_to_multispectrum()

        self.spectrum.exclude = prev_exclude  # make the exclude is reset
        multi_back = self.spectrum.copy()
        multi_back.multidata = self.spectrum.multidata \
            - self.get_background_model().multidata
        multi_back.setcurrentspectrum((0, 0))

        self.result = multi_back
        return self.result

    def get_background_model(self):
        return self.multi_model_signal

    def get_fitting_parameters(self):
        """
        Returns the fitting parameters
        :return:
        """
        return self.fitter.coeff_matrix

    def fast_calculate_multi(self):
        """
        Uses the parameters estimated from the autofit to remove the background
        this only works if the model_type is Powerlaw or Exponential

        Returns
        -------
        self.fast_result: MultiSpectrum
            The background subtracted multispectrum using the fast fitting
            parameters.

        """

        prev_exclude = self.spectrum.exclude[:]
        self.set_indices()
        self.include_areas()

        if self.multi_fast_fit is None:
            logger.info('Determination of fast fit paramters'
                        ' is peformed first')
            self.determine_multi_fast_fit_parameters()

        fast_back = self.spectrum.copy()  # background subtracted
        rem_sig = self.spectrum.copy()  # calculated background
        ndata = np.empty((self.spectrum.xsize,
                          self.spectrum.ysize, self.spectrum.size))
        rem_dat = np.empty(ndata.shape)
        for index in tqdm(np.ndindex(fast_back.multidata.shape[:-1])):
            islice = np.s_[index]

            A = self.multi_fast_fit[islice+(0,)]
            r = self.multi_fast_fit[islice+(1,)]

            self.model.getfreeparameters()[0].setvalue(A)
            self.model.getfreeparameters()[1].setvalue(r)
            self.model.calculate()
            ndata[islice] = self.spectrum.multidata[islice] - self.model.data
            rem_dat[islice] = self.model.data

        fast_back.multidata = ndata
        rem_sig.multidata = rem_dat
        self.spectrum.exclude = prev_exclude  # make the exclude is reset
        self.multi_model_signal = rem_sig
        self.fast_result = fast_back
        return fast_back

    def show_fit_result(self, index=(0, 0), integration_range=None,
                        use_mean=False):
        """
        Shows the background fitting result. Is usefull for showing the result
        of the procedure to identify problems or being able to optimize the
        signal_range.

        Parameters
        ----------
        index: tuple
            If a multispectrum is used, then the index inputted will be used
            for the visualization
        integration_range: tuple
            Input the expected integration range to get the elemental
            abundance. Just used for visualization purpose.
        use_mean: boolean
            If a multispectrum is used, the average results can be shown which
            could be useful when wanting to decrease the noise level in the
            individual spectrum.

        Returns
        -------
        fig: Figure
            The created figure
        """
        prev_exclude = self.spectrum.exclude[:]
        self.set_indices()
        self.include_areas()

        if isinstance(self.spectrum, MultiSpectrum):
            self.fitter.setcurrentfit(index)

        fig, ax = plt.subplots()
        if use_mean:
            ax.plot(self.spectrum.energy_axis, self.spectrum.mean().data,
                    color='black', label='Data')
            ax.plot(self.spectrum.energy_axis,
                    self.multi_model_signal.mean().data, color='red',
                    label='Background model')

        else:
            ax.plot(self.spectrum.energy_axis, self.spectrum.data,
                    color='black', label='Data')
            ax.plot(self.spectrum.energy_axis, self.model.data,
                    color='red', label='Background model')

        ax.fill_between(self.spectrum.energy_axis, 0,
                        1.1*self.spectrum.data.max(),
                        where=np.invert(self.spectrum.exclude), color='green',
                        alpha=0.5, label='Background region')
        ax.set_xlabel(r'Energy Loss [eV]')
        ax.set_ylabel(r'Counts')
        if integration_range is not None:
            boolean = (integration_range[0] <= self.spectrum.energy_axis) \
                      & (integration_range[1] >= self.spectrum.energy_axis)
            ax.fill_between(self.spectrum.energy_axis, self.model.data,
                            self.spectrum.data, where=boolean,
                            color='blue', alpha=0.5,
                            label='Integration region')
        ax.legend()
        ax.set_ylim([0, 1.1*self.spectrum.data.max()])
        if isinstance(self.spectrum, MultiSpectrum):
            ax.set_title(self.spectrum.currentspectrumid)

        self.spectrum.exclude = prev_exclude  # make the exclude is reset
        return fig

    def quantify_from_edge(self, integration_range, element, edge, E0, alpha,
                           beta, ll=None):
        """
        Calculates the abundance of the element using the background removed
        edge. The integration window should be added together with information
        on which element, and the experimental parameters. A low loss could
        be provided to get a more accurate result.


        Parameters
        ----------
        integration_range: tuple
            Input the expected integration range to get the elemental
            abundance. Just used for visualization purpose.
        element: string
            The element of interest
        edge: string
            The edge being used
        E0: float
            acceleration voltage [V]
        alpha: float
            convergence angle [rad]
        beta: float
            collection angle [rad]
        ll: Spectrum or MultiSpectrum
            The low loss of the used core-loss spectrum which is used to
            convolve the theoretical core-loss edges and have a more accurate
            result.

        Returns
        -------
        abundance: float, numpy 2d array, None
            The calculated abundance using the background removal method.
            A float is returned for a spectrum and a 2d numpy array when a
            multispectrum is used. None is returned when this function is
            called before any background removal has been applied.

        """

        specshape = self.spectrum.get_spectrumshape()
        comp_elements = []
        comp_elements.append(ZezhongCoreLossEdgeCombined(specshape, 1, E0,
                                                         alpha, beta, element,
                                                         edge))

        if isinstance(ll, MultiSpectrum):
            shape = (self.spectrum.xsize, self.spectrum.ysize)
            calc_int = np.zeros(shape)

            for index in tqdm(np.ndindex(shape)):
                islice = np.s_[index]
                ll0 = ll[islice[0], islice[1], :]
                llcomp = MscatterFFT(specshape, ll0)
                mod = Model(specshape, components=comp_elements+[llcomp])
                mod.calculate()
                calc_int[islice] = mod.integrate(integration_range)

        else:
            if ll is not None:
                comp_elements.append(MscatterFFT(specshape, ll))

            mod = Model(specshape, components=comp_elements)
            mod.calculate()

            if isinstance(self.spectrum, MultiSpectrum):
                shape = (self.spectrum.xsize, self.spectrum.ysize)
                calc_int = np.zeros(shape)
                calc_int[:, :] = mod.integrate(integration_range)
            else:
                calc_int = mod.integrate(integration_range)

            if self.result is None:
                print('uses the fast calculated background fit')
                return None

        exp_int = self.result.integrate(integration_range)
        abundance = exp_int / calc_int
        return abundance



# the classes in the cores
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.model import Model

#the classes from the components
from pyEELSMODEL.components.CLedge.zezhong_coreloss_edgecombined \
    import ZezhongCoreLossEdgeCombined
from pyEELSMODEL.components.CLedge.kohl_coreloss_edgecombined \
    import KohlLossEdgeCombined
from pyEELSMODEL.components.CLedge.hydrogen_coreloss_edge \
    import HydrogenicCoreLossEdge
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT

#the fitters, mainly linear and least squares optimization
from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.fitters.lsqfitter import LSQFitter
from pyEELSMODEL.fitters.quadratic_fitter import QuadraticFitter

#operators which are useful
from pyEELSMODEL.operators.backgroundremoval import BackgroundRemoval
from pyEELSMODEL.operators.aligns.fastalignzeroloss import FastAlignZeroLoss
from pyEELSMODEL.operators.aligns.alignzeroloss import AlignZeroLoss
from pyEELSMODEL.operators.aligns.aligncrosscorrelation import\
    AlignCrossCorrelation
from pyEELSMODEL.operators.multispectrumvisualizer import \
    MultiSpectrumVisualizer
from pyEELSMODEL.operators.areaselection import \
    AreaSelection
from pyEELSMODEL.operators.calibratespectrum import \
    CalibrateSpectrum



from pyEELSMODEL.operators.simulator.coreloss_simulator import\
    CoreLossSimulator
from pyEELSMODEL.operators.quantification.elemental_quantification import\
    ElementalQuantification




"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
from scipy.optimize import nnls

from pyEELSMODEL.core.fitter import Fitter
from pyEELSMODEL.components.MScatter.mscatter import Mscatter


class LinearFitter(Fitter):
    """
    A linear fitter class which uses as input a Spectrum or Multispectrum and
    Model.
    """

    def __init__(self, spectrum, model, method='ols', use_weights=False):

        """
        Initialises a LinearFitter instance. Two different methods can be
        used which is the ordinary least squares (ols) or the non-negative
        least squares (nnls). The nnls makes that the values of the fitted
        parameters cannot become negative.

        Parameters
        ----------
        spectrum : Spectrum or Multispectrum
            The experimental data used which needs to be fitted
        model : Model
            The model used to fit the experimental data. Model should only
            contain linear parameters.
        method: str, optional
            String indicating which linear fitting procedure should be used. .
            Available fitters are: ols and nnls
        use_weights: boolean
            Indicating if the weights are being used during the fit. The
            weigths assume poisson noise.

        Returns
        -------
        An instance of a LinearFitter.

        """
        # model can only be check if it is add
        if not model.islinear():
            raise ValueError(r'There are non-linear parameters in the model')

        super().__init__(spectrum, model)

        self.method = method
        self.A_matrix = None
        self.y_vector = None
        # a list where each component of the A matrix column is saved
        self.component_list = []
        self.same_A_matrix = True  # use the same A matrix all the time
        self.use_weights = use_weights

    @property
    def A_matrix(self):
        return self._A_matrix

    @A_matrix.setter
    def A_matrix(self, A):
        self._A_matrix = A

    @property
    def y_vector(self):
        return self._y_vector

    @y_vector.setter
    def y_vector(self, y):
        self._y_vector = y

    def calculate_A_matrix(self):
        """
        Get derivative matrix. This is a matrix which holds for each energy in
        the spectrum the first derivative of the value at that energy with
        respect to each free parameter in the model.
        This A matrix needs to be convolved to include the low loss scattering

        """
        ncol = self.model.getnumfreeparameters()
        A = np.zeros((self.spectrum.size, ncol))
        component_list = []

        for i, param in enumerate(self.model.getfreelinparameters()):
            # put all parameters to 0 except 1
            for par in self.model.getfreelinparameters():
                if par == param:
                    # this could cause trouble as it ignores any scaling issues
                    par.setvalue(1)
                else:
                    par.setvalue(0)
            self.model.calculate(use_ll=False)
            component_list.append(self.model.getcomponentbyparameter(param))
            A[:, i] = self.model.data

        self.A_matrix = A
        self.component_list = component_list

    def convolute_A_matrix(self):
        """
        Convolutes the A matrix with the low loss. This is necessary to
        take the low loss into account. It does this by first checking which
        components need to be convolved. The next step is to convolve each
        row of the A matrix with low loss.
        This low loss convolution increases the fitting time. Other and faster
        ways can be exploited to have a faster quantification.

        Returns
        -------
        A_matrix_ll: numpy 2d array
            The convolved A matrix with the low loss.

        """

        for comp in self.model.components:
            if isinstance(comp, Mscatter):
                ll_comp = comp

        A_matrix_ll = np.empty(self.A_matrix.shape)
        ll_comp.new_ll = True
        for i in range(self.A_matrix.shape[1]):
            if self.component_list[i].getcanconvolute():
                ndata = self.A_matrix[:, i]
                ll_comp.data = ndata
                ll_comp.calculate()
                ll_comp.new_ll = False
                A_matrix_ll[:, i] = ll_comp.data
            else:
                A_matrix_ll[:, i] = self.A_matrix[:, i]
        ll_comp.new_ll = True
        return A_matrix_ll

    def convolute_A_matrix_(self):
        """
        Test convolution using the matrix approach however this seems
        to be slower than the original which loops over A matrix
        """
        for comp in self.model.components:
            if isinstance(comp, Mscatter):
                ll_comp = comp

        A_matrix_ll = ll_comp.calculate_A_matrix(self.A_matrix)
        for i in range(self.A_matrix.shape[1]):
            if not self.component_list[i].getcanconvolute():
                A_matrix_ll[:, i] = self.A_matrix[:, i]
        return A_matrix_ll

    def calculate_y_vector(self):
        """
        Calculates the y vector for the linear least squares fitting.
        The y vector is the raw data where the excluded regions are not
        included.

        Returns
        -------
        y: numpy array
            The y vector for the linear least squares fitting

        """
        y = self.spectrum.data[np.invert(self.spectrum.exclude)]
        self.y_vector = y

    def calculate_weights(self):
        """
        The weights are calculated assuming the data is poisson noise.
         W=1/(sigma**2), and sigma**2 = N which is the number of electrons.
         If the data itself is lower than 1, which could happen due to dark
         noise, then the value is set to 1. Note that the data itself needs
         to be expressed in electron (right pppc) for it to make most sense.

         Returns
        -------
        weights: numpy array
            The weigths used in the fitting procedure for poisson noise.

        """
        ndata = np.copy(self.y_vector)
        ndata[ndata < 1] = 1
        weigths = 1/ndata
        return weigths

    def perform_fit(self):
        """
        Fits the spectrum using the linear least squares method.
        """

        if self.A_matrix is None or \
                self.model.ischanged() or \
                not self.same_A_matrix:
            self.calculate_A_matrix()  # makes it slow
            self.model.setchanged(False)

        if self.model.hasconvolutor():
            A = self.convolute_A_matrix()[np.invert(self.spectrum.exclude), :]
        else:
            A = self.A_matrix[np.invert(self.spectrum.exclude), :]

        self.calculate_y_vector()
        y = self.y_vector

        if self.use_weights:
            weights = self.calculate_weights()
            tmp = np.sqrt(weights)
            AW = A * tmp[:, np.newaxis]
            yW = y * tmp
        else:
            AW = A
            yW = y

        if self.method == 'nnls':
            try:
                coeff, cost = nnls(AW, yW)
                self.coeff = coeff
                self.error = cost
            except Exception:
                self.coeff = np.nan
                self.error = np.nan

        elif self.method == 'ols':
            coeff, resi, rank, sing = np.linalg.lstsq(AW, yW, rcond=-1)
            self.coeff = coeff
            if resi.size == 0:
                self.error = np.inf
            else:
                self.error = resi[0]

        else:
            print('Fitting method is wrong')

    def pearson_correlation_matrix(self):
        """
        Calculates the pearson correlation matrix
        Returns
        -------
        correlation_matrix: 2d numpy array
            The pearson correlation matrix
        """
        self.calculate_A_matrix()
        A = self.A_matrix

        correlation_matrix = np.zeros((A.shape[1], A.shape[1]))
        for i in range(A.shape[1]):
            for j in range(A.shape[1]):
                r = stats.pearsonr(A[:, i], A[:, j])[0]
                correlation_matrix[i, j] = r

        return correlation_matrix

    def show_pearson_correlation(self):
        correlation = self.pearson_correlation_matrix()
        label_list = self.getlabelist()
        fig, ax = plt.subplots()
        ax.imshow(correlation, vmin=-1, vmax=1, cmap='bwr')
        ax.set_xticks(np.arange(len(label_list)))
        ax.set_yticks(np.arange(len(label_list)))
        ax.set_xticklabels(label_list, rotation=45, ha="right")
        ax.set_yticklabels(label_list, rotation=45)
        fig.set_tight_layout(True)

    def partial_derivative(self, parameter):
        """
        Calculates the partial derivative of the parameter. When the entire
        model is linear. This is taking the rows (or columns) of the A matrix
        which needs to be convoluted when low loss is used.

        Parameters
        ----------
        parameter: Parameter
            The parameter from which the partial derivative will be
            calculated.
        fraction: float
            The fraction of the step the parameter does to calculate
            the numerical gradient. (default: 0.001)

        Returns
        -------
        parial: numpy array (1D)
            The partial derivative for the given parameter.
        """

        if self.model.hasconvolutor():
            A_matrix = self.convolute_A_matrix()
        else:
            A_matrix = self.A_matrix

        index = self.get_param_index(parameter)
        return A_matrix[np.invert(self.spectrum.exclude), index]



from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge

from pyEELSMODEL.components.CLedge.zezhong_coreloss_edge import \
    ZezhongCoreLossEdge
from pyEELSMODEL.database.Zhang.download import download_file
import numpy as np
import logging
import h5py
import os
from pyEELSMODEL import __file__

logger = logging.getLogger(__name__)


# todo implement method

class ZezhongCoreLossEdgeCombined(CoreLossEdge):
    """
    Calculates the coreloss edges for a group of edges using the GOS from
    Zezhong Zang. For instance, the L edge calculates the L3, L2 and L1 edges
    and puts them together with the appropriate prefactors.

    """

    def __init__(
            self,
            specshape,
            A,
            E0,
            alpha,
            beta,
            element,
            edge,
            eshift=0,
            q_steps=100,
            dir_path=None,
    ):
        """
        Coreloss edges which are calculated by Zezhong Zhang.
        https://zenodo.org/records/11199911


        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
              Amplitude of the edge

        E0: float [V]
              The acceleration voltage of the microscope

        alpha: float [rad]
              The convergence angle of the incoming probe

        beta: float [rad]
              The collection angle

        element: string
              The name of the element from which to calculate the edge model.

        edge: string
              The type of edge. (K, L or M)

        eshift: float [eV]
              The shift of the onset energy with respect to the literature
              value. (default: 0)

        q_steps: int
             The number of q points taken into account for the integration over
             the momentum space. The larger the number of q_steps the more
             accurate the calculation. (default: 100)

        Returns
        -------

        """

        if dir_path is None:
            self.dir_path = os.path.dirname(
                os.path.dirname(__file__) + "/../pyEELSMODEL/database/Zhang/"
            )
        else:
            self.dir_path = dir_path
        self.file = os.path.join(self.dir_path, "Dirac_GOS.gosh")
        if not os.path.exists(self.file):
            download_file(filename=self.file)

        self.xsectionlist = []
        max_edge = self.check_maximum_edge(specshape, element, edge)

        if edge == "K":
            xsectionK = ZezhongCoreLossEdge(
                specshape,
                A,
                E0,
                alpha,
                beta,
                element,
                "K1",
                eshift=eshift,
                q_steps=q_steps,
                dir_path=self.dir_path,
            )
            self.xsectionlist.append(xsectionK)
            super().__init__(specshape, A, E0, alpha, beta, element, "K1",
                             q_steps=q_steps)
            name = element + " K edge: " + str(self.onset_energy) + " eV"
            self.setdisplayname(name)

        elif (edge == "L") | (edge == "M") | (edge == "N") | (edge == "O"):
            start_edge = edge + str(max_edge)
            xsection_op = ZezhongCoreLossEdge(
                specshape,
                A,
                E0,
                alpha,
                beta,
                element,
                start_edge,
                eshift=eshift,
                q_steps=q_steps,
                dir_path=self.dir_path,
            )
            self.xsectionlist.append(xsection_op)
            for i in range(max_edge - 1):
                try:
                    next_edge = edge + str(max_edge - i - 1)
                    xsection = ZezhongCoreLossEdge(
                        specshape,
                        A,
                        E0,
                        alpha,
                        beta,
                        element,
                        next_edge,
                        eshift=eshift,
                        q_steps=q_steps,
                        dir_path=self.dir_path,
                    )
                    xsection.parameters[0].couple(xsection_op.parameters[0])
                    xsection.parameters[1].couple(xsection_op.parameters[1])
                    xsection.parameters[2].couple(xsection_op.parameters[2])
                    xsection.parameters[3].couple(xsection_op.parameters[3])
                    self.xsectionlist.append(xsection)
                    # print(next_edge + " is used")

                except Exception:
                    print(next_edge + " is NOT implemented")

            super().__init__(
                specshape,
                A,
                E0,
                alpha,
                beta,
                element,
                start_edge,
                eshift=eshift,
                q_steps=100,
            )
            name = element + edge + " edge: " + str(self.onset_energy) + " eV"
            self.setdisplayname(name)

        self.manageparameters()
        self.calculate()

    def check_maximum_edge(self, specshape, element, edge):
        """
        Check which is the lowest energy loss edge (the highest number)
        available. Some elements only have a L1 but do not have a L3 edge.
        This function can tell which number is the highest for the given
        element and edge.

        Parameters
        ----------
        element : string
            The element from which the information should be retrieved.
        edge : string
            The edge from which the information should be retrieved.

        Returns
        -------
        max_value: uint > 0
            The integer from the edge having the lowest energy onset.

        """
        with h5py.File(self.file, "r") as f:
            # print('file is opened')
            edges = list(f[element].keys())
            onset_energy_list = list(
                f[element][i]["metadata"].attrs["onset_energy_guess"]
                for i in edges
            )
        sub_edges = []
        sub_onsets = []
        for char in edges:
            if edge in char:
                sub_edges.append(int(char[-1]))
                sub_onsets.append(onset_energy_list[edges.index(char)])
        sub_onsets = np.array(sub_onsets)
        sub_edges = np.array(sub_edges)

        begin = specshape.offset
        end = specshape.offset + specshape.size * specshape.dispersion
        mask = np.logical_and((sub_onsets > begin), (sub_onsets < end))
        # mask = (sub_onsets>begin) & (sub_onsets<end)
        mask = np.ravel(mask)
        # sub_edges = np.array(sub_edges)[mask]

        av_e = np.array(sub_edges)[mask]
        co = np.argsort(av_e)[::-1]

        # stores all the edges used in the calculation of the components
        # self.available_edges = av_e[co]
        # self.available_onsets = np.array(sub_onsets)[mask].flatten()[co]

        return max(av_e[co])

    def manageparameters(self):
        self.parameters = (
            []
        )  # erase existing parameter list and replace with longer list
        self.gradient = []
        # first 4 paramters however remain same meaning
        # but they are coupled with the higher ones
        for xsection in self.xsectionlist:
            for par in xsection.parameters:
                self._addparameter(par)

    def set_onset_energy(self):
        """
        Sets the onset energy of the grouped edge.
        """
        self.onset_energy = self.xsectionlist[
            0
        ].onset_energy  # first part of cross section determines onset

    def set_edge(self, edge):
        self.edge = self.xsectionlist[0].edge

    def calculate_cross_section(self):
        """
        Calculates the cross section of the combined edge
        """
        cross_section = np.zeros(self.size)
        for xsection in self.xsectionlist:
            if xsection.suppress==False:
                cross_section += xsection.calculate_cross_section()
            else:
                print('subcomponent is suppressed, to unset call setsuppress(False) on the specific subcomponent:',xsection.name)            
        return cross_section

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        pA = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        p4 = self.parameters[3]

        if p2.ischanged() or p3.ischanged() or p4.ischanged():
            self.cross_section = self.calculate_cross_section()
            self.data = pA.getvalue() * self.cross_section

        if pA.ischanged():
            self.data = pA.getvalue() * self.cross_section
        self.setunchanged()



"""
Testing the functionality of the zero loss alignment
"""

import numpy as np
import pytest
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.spectrum import Spectrumshape
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.lorentzian import Lorentzian
from pyEELSMODEL.operators.aligns.alignzeroloss import AlignZeroLoss
from pyEELSMODEL.operators.aligns.fastalignzeroloss import FastAlignZeroLoss
from pyEELSMODEL.operators.aligns.aligncrosscorrelation import \
    AlignCrossCorrelation

# Make a dataset where the centre changes.
# Every column has a incremental increase and between every row there is a
# random offset


def make_artificial_dataset(start, end, model_type="Gaussian",
                            return_shift=False):
    """
    Make artificial datasets used to test the performance of the zero loss
    alignment.
    """

    ncol = 100  # number of columns
    nrow = 12  # number of rows
    width = 10  # widht of the centre misalignement
    fwhm = 3  # fwhm of gaussian or lorentzian
    amp = 100  # amplitude
    specshape = Spectrumshape(0.5, -200, 1024)

    centre_array = np.linspace(start * width, end * width, ncol)
    offs_array = 10 * np.random.rand(nrow)
    offs_array = 10 * np.ones(nrow)
    sim_data = np.zeros((nrow, ncol, specshape.size))
    sim1_data = np.zeros(
        sim_data.shape)  # extra spectra which will be aligned using zl
    sim2_data = np.zeros(
        sim_data.shape)  # extra spectra which will be aligned using zl
    real_shift = np.zeros((nrow, ncol))
    for i in range(nrow):
        for j in range(ncol):
            if model_type == 'Gaussian':
                zl = Gaussian(specshape, amp, centre_array[j] + offs_array[i],
                              fwhm)
                zl1 = Gaussian(specshape, amp,
                               50 + centre_array[j] + offs_array[i], fwhm)
                zl2 = Gaussian(specshape, amp,
                               100 + centre_array[j] + offs_array[i], fwhm)

            else:
                zl = Lorentzian(specshape, amp,
                                centre_array[j] + offs_array[i], fwhm)
                zl1 = Lorentzian(specshape, amp,
                                 50 + centre_array[j] + offs_array[i], fwhm)
                zl2 = Lorentzian(specshape, amp,
                                 100 + centre_array[j] + offs_array[i], fwhm)

            zl.calculate()
            zl1.calculate()
            zl2.calculate()

            # sim_data[i, j] = np.random.poisson(zl.data)
            # sim1_data[i, j] = np.random.poisson(zl1.data)
            # sim2_data[i, j] = np.random.poisson(zl2.data)

            sim_data[i, j] = zl.data
            sim1_data[i, j] = zl1.data
            sim2_data[i, j] = zl2.data

            real_shift[i, j] = centre_array[j] + offs_array[i]

    mshape = MultiSpectrumshape(specshape.dispersion, specshape.offset,
                                specshape.size, nrow, ncol)
    s = MultiSpectrum(mshape, data=sim_data)
    s1 = MultiSpectrum(mshape, data=sim1_data)
    s2 = MultiSpectrum(mshape, data=sim2_data)
    if return_shift:
        return s, s1, s2, real_shift
    else:
        return s, s1, s2


def test_init_other():
    s, _, _ = make_artificial_dataset(-2, 2)

    # #wrong xsize
    mshape = MultiSpectrumshape(1, 10, 1024, s.xsize + 20, s.ysize)
    s1 = MultiSpectrum(mshape)
    with pytest.raises(ValueError):
        AlignZeroLoss(s, other_spectra=[s1])

    # #wrong ysize
    mshape = MultiSpectrumshape(1, 10, 1024, s.xsize, s.ysize + 20)
    s1 = MultiSpectrum(mshape)
    with pytest.raises(ValueError):
        AlignZeroLoss(s, other_spectra=[s1])


def test_fast_align():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = FastAlignZeroLoss(s)
        align.perform_alignment()

        p = align.aligned

        # every maximum value is at zero energy
        assert np.all(p.energy_axis[np.argmax(p.multidata, axis=2)] == 0)


def test_fast_align_other():
    s, s1, s2 = make_artificial_dataset(-2, 2)
    align = FastAlignZeroLoss(s, other_spectra=[s1, s2])
    align.perform_alignment()

    p0 = align.aligned_others[0]
    p1 = align.aligned_others[1]

    assert p0.energy_axis[np.argmax(p0.sum().data)] == 50
    assert p1.energy_axis[np.argmax(p1.sum().data)] == 100


def test_fast_align_crop():
    shf_l = [[-2, 2], [-4, -2], [1, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = FastAlignZeroLoss(s, other_spectra=[s1, s2], cropping=True)
        align.perform_alignment()

        new_size = s.size - int(
            (real_shift.max() - real_shift.min()) / s.dispersion)

        # print(new_size)
        # print(align.aligned.size)

        assert align.aligned.size == new_size
        assert align.aligned_others[0].size == new_size
        assert align.aligned_others[1].size == new_size


def test_fit_align():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1])
        align = AlignZeroLoss(s)
        align.perform_alignment()

        # every maximum value is almost at zero
        assert np.all(np.abs(
            s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)


def test_fit_align_lorentz():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2 = make_artificial_dataset(shf_l[i][0], shf_l[i][1],
                                            model_type='Lorentzian')
        align = AlignZeroLoss(s, model_type='Lorentzian')
        align.perform_alignment()

        # every maximum value is almost at zero
        assert np.all(np.abs(
            s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)


def test_fit_align_other():
    s, s1, s2 = make_artificial_dataset(-2, 2)
    align = AlignZeroLoss(s, other_spectra=[s1, s2])
    align.perform_alignment()

    # every maximum value is almost at zero
    assert np.all(
        np.abs(s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]) < 2)
    assert np.all(np.abs((s1.energy_axis[
        np.argmax(align.aligned_others[0].multidata, axis=2)]) - 50) < 2)
    assert np.all(np.abs((s2.energy_axis[
        np.argmax(align.aligned_others[1].multidata, axis=2)]) - 100) < 2)


def test_fit_align_crop():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = AlignZeroLoss(s, other_spectra=[s1, s2], cropping=True)
        align.perform_alignment()

        new_size = s.size - int(
            np.abs(real_shift.max() - real_shift.min()) / s.dispersion)

        print(new_size)
        print(align.aligned.size)

        assert align.aligned.size == new_size
        assert align.aligned_others[0].size == new_size
        assert align.aligned_others[1].size == new_size


def test_align_crosscor():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = AlignCrossCorrelation(s, other_spectra=[s1, s2], cropping=True)
        align.perform_alignment()

        max_en = s.energy_axis[np.argmax(align.aligned.multidata, axis=2)]

        assert np.all(np.isclose(max_en, max_en[0, 0]))


def test_align_crosscor_zlp():
    shf_l = [[-2, 2], [-4, -2], [2, 4]]
    for i in range(len(shf_l)):
        s, s1, s2, real_shift = make_artificial_dataset(shf_l[i][0],
                                                        shf_l[i][1],
                                                        return_shift=True)
        align = AlignCrossCorrelation(s, other_spectra=[s1, s2], cropping=True,
                                      is_zlp=True)
        align.perform_alignment()

        max_en = align.aligned.energy_axis[
            np.argmax(align.aligned.multidata, axis=2)]
        max_en1 = align.aligned_others[0].energy_axis[
            np.argmax(align.aligned_others[0].multidata, axis=2)]
        max_en2 = align.aligned_others[1].energy_axis[
            np.argmax(align.aligned_others[1].multidata, axis=2)]

        assert np.all(np.isclose(max_en, 0))
        assert np.all(np.isclose(max_en1, 50))
        assert np.all(np.isclose(max_en2, 100))


def main():
    test_init_other()
    test_fast_align()
    test_fast_align_other()
    test_fast_align_crop()
    test_fit_align()
    test_fit_align_lorentz()
    test_fit_align_other()
    test_fit_align_crop()
    test_align_crosscor()
    test_align_crosscor_zlp()


if __name__ == "__main__":
    main()



import numpy as np
import pyEELSMODEL.api as em


# functions to make easy mask to make elemental maps
def make_circular_mask(xco, yco, Rin, Rout, shape):
    XX, YY = np.meshgrid(np.arange(shape[1]) - yco, np.arange(shape[0]) - xco)
    R = np.sqrt(XX ** 2 + YY ** 2)

    mask = np.zeros(shape)
    boolean = (R >= Rin) & (R < Rout)
    mask[boolean] = 1
    return mask


def make_rectangular_mask(xco, yco, width, height, shape):
    mask = np.zeros(shape)
    mask[xco:xco + width, yco:yco + height] = 1
    return mask


def simulate_data():
    """
    Small function which simulates a typical STEM-EELS map.

    """
    elements = ['C', 'N', 'O', 'Fe']
    edges = ['K', 'K', 'K', 'L']
    Zs = [6, 7, 8, 26]  # atomic weights

    # scan size elemental map
    xsize = 128
    ysize = 128
    maps = np.zeros((len(elements), xsize, ysize))

    # masks which define different regions.
    mask0 = make_rectangular_mask(5, 5, 20, 20, (xsize, ysize))
    mask1 = make_rectangular_mask(90, 90, 20, 30, (xsize, ysize))
    mask2 = make_circular_mask(xsize // 2, ysize // 2, 20, 30, (xsize, ysize))
    mask3 = make_circular_mask(xsize // 2, ysize // 2, 0, 20, (xsize, ysize))

    # attribute different elemental values for different masks
    maps[0] = 1  # carbon elemental map
    maps[1] = 2 * mask0 + mask1  # nitrogen elemental map
    maps[2] = mask2  # oxygen elemental map
    maps[3] = mask3 + 0.5 * mask2  # iron elemental map

    adf = np.zeros((xsize, ysize))
    tlambda_map = np.zeros_like(adf)
    for ii in range(maps.shape[0]):
        adf += (Zs[ii] * maps[ii]) ** 2
        tlambda_map += Zs[ii] * maps[ii]

    # maximum t/lambda is 1 for this sample
    tlambda_map = tlambda_map / tlambda_map.max()
    E0 = 300e3  # acceleration voltage [V]
    alpha = 1e-9  # convergence angle [rad]
    beta = 20e-3  # collection angle [rad]

    dispersion = 0.5  # [eV]
    offset = 200  # [eV]
    size = 2048  # number of pixel used in simulation [eV]

    settings = (E0, alpha, beta)
    msh = em.MultiSpectrumshape(dispersion, offset, size, xsize, ysize)

    sim = em.CoreLossSimulator(msh, elements, edges, maps, tlambda_map,
                               settings)
    # add shift which experimentally arises due to instabilities
    sim.use_shift = True
    sim.simulate_multispectrum()

    hl = sim.multispectrum
    ll = sim.ll
    ll.multidata = ll.multidata * 1e4
    return hl, ll



from fpdf import FPDF
import time
import os


class PDF(FPDF):
    """
    Class which makes a pdf, this is useful when wanting to have some
    information used on a procedure you applied to the data. For instance,
    the FastAlignZeroLoss class can generate a file with the relevant
    information which can be checked if everything went fine.
    """

    def get_time_name(self):
        return time.ctime(time.time()).replace(" ", "_").replace(":", "_")

    def get_savename(self):
        name = self.get_time_name() + '_shift_overview.pdf'
        return os.path.join(self.savepath, name)

    def get_figname(self):
        name = self.get_time_name() + '_img' + str(self.counter) + '.png'
        self.counter += 1
        return os.path.join(self.savepath, name)

    def add_figure(self, fig):
        savename = self.get_figname()
        fig.savefig(savename, bbox_inches='tight', dpi=300)
        wd = fig.get_figwidth()
        he = fig.get_figheight()
        self.set_xy(0.0, self.get_y() + self.dh)

        self.image(savename, link='', type='', w=wd, h=he)
        os.remove(savename)

    def intitialize(self, path):
        self.set_font('Arial', size=12)
        self.counter = 0
        self.dh = 0.2
        self.savepath = path



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import logging

from pyEELSMODEL.core.model import Model

from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.lorentzian import Lorentzian

from pyEELSMODEL.fitters.lsqfitter import LSQFitter
from pyEELSMODEL.operators.aligns.align import Align

logger = logging.getLogger(__name__)


class AlignZeroLoss(Align):
    """
    AlignZeroLoss is a class which aligns the zero loss peak and other spectra
    can be added which also need to be aligned.
    The zero loss is fitted to a model (Gaussian or Lorentzian) and this
    parameter will be used to align the zero loss. Note that some interpolation
    step is used to shift the data. This could introduce differences
    in noise properties when wanting to perform the most accurate statistics
    on it.
    """
    def __init__(self, multispectrum, other_spectra=None,
                 signal_range=None, model_type='Gaussian', cropping=False,
                 use_bounds=True):
        """

        Parameters
        ----------
        multispectrum: Multispectrum
            The spectrum from which the background should be removed
        other_spectra: List of Multispectra
            The other multispectra which can be aligned with the same
            parameters. (default: None)
        signal_range: tuple
            Indicates the region on which the zero loss fitting should be
            performed.
            (default: None)
        model_type: string
            Selecting which model is used to fit the zero loss peak
             (default: 'Gaussian')
        cropping: bool
            Indicates if the edges of the signal are cropped out when aligning
            the zero loss peak.
            (default: False)
        use_bounds: bool
            Use the boundaries for the fitting. This could make the fit more
            stable.
            (default: True)
        """

        super().__init__(multispectrum, other_spectra, cropping,
                         signal_range=signal_range)

        self.model_type = model_type
        self.use_bounds = use_bounds

        self.aligned = None  # attribute which stores the aligned multispectrum
        self.aligned_others = []  # attribute
        self.start_parameters = None
        self._fitter = None
        self.model = None
        self.make_zeroloss_model()
        self.method = 'Model based fitting'

    @property
    def model(self):
        """
        The model used for the fitting
        """
        return self._model

    @model.setter
    def model(self, m0):
        """
        Sets the attribute model to the given model.
        """
        self._model = m0

    @property
    def fitter(self):
        """
        The fitter used in the spectrum (now it is the LSQ since it fast
        and stable)
        """
        return self._fitter

    @fitter.setter
    def fitter(self, fit):
        self._fitter = fit

    def estimate_start_parameters(self):
        """
        Estimates the starting values for every scan position to fit the zero
        loss peak. The A value is found by taking the maximum.
        The centre is chosen as the coordinate at which the maximum occurs
        Sigma is taken as a fixed value of 1. (better guess can be tried but it
         seems to be stable)
        It stores the result is the start_parameters attribute.
        """
        # The lorentzian and gaussian model both have three parameters.
        start_param = \
            np.zeros((self.multispectrum.xsize, self.multispectrum.ysize, 3))

        start_param[:, :, 0] = np.max(self.multispectrum.multidata, axis=2)

        # this should be zero in principle
        ind0 = self.multispectrum.get_energy_index(self.signal_range[0])
        ind1 = self.multispectrum.get_energy_index(self.signal_range[1])
        co = np.argmax(self.multispectrum.multidata[:, :, ind0:ind1], axis=2)
        shift = (co + ind0)
        start_param[:, :, 1] = self.multispectrum.energy_axis[shift]

        # fwhm is guess from the width of the summed spectrum.
        # this says that the width does not change over the scan
        # (which is normally true)
        # zl = self.fast_aligned.multidata.sum((0,1))
        # coor = np.argwhere(zl>0.5*np.max(zl))

        start_param[:, :, 2] = 1

        self.start_parameters = start_param

    def set_indices(self):
        """
        Calculates the indices used which are excluded in the fit.
        These indices are also used to determine a first guess of the
        background model. The result is stored in the indices attribute

        """

        ind1 = [self.multispectrum.get_energy_index(self.signal_range[0]),
                self.multispectrum.get_energy_index(self.signal_range[1])]
        self.indices = ind1

    def make_zeroloss_model(self):
        """
        Creates a model for the zero loss peak, this depends on which
        model_type is chosen when creating the background object. The model
        is stored in the model attribute
        """
        specshape = self.multispectrum.get_spectrumshape()
        m0 = Model(specshape)
        if self.model_type == 'Gaussian':
            comp = Gaussian(specshape, A=1, centre=0, fwhm=1)
        elif self.model_type == 'Lorentzian':
            comp = Lorentzian(specshape, A=1, centre=0, fwhm=1)
        else:
            print('model type not included in the list of possibilities')

        comp.parameters[1].setboundaries(self.signal_range[0],
                                         self.signal_range[1], force=True)

        m0.addcomponent(comp)
        self.model = m0

    def include_areas(self):
        """
        Sets the exlude of the spectrum such that only the integration
        range is taken into account.
        """
        self.multispectrum.set_include_region(self.indices[0], self.indices[1])

    def calculate_model(self):
        """
        Fits the model to the data. The fitting region is resetted as it
        initially was after the fit.
        """

        prev_exclude = self.multispectrum.exclude[:]
        self.multispectrum.exclude = \
            np.ones(self.multispectrum.size, dtype=bool)
        self.set_indices()
        self.include_areas()

        # fit = MLFitter(self.spectrum, self.get_model())
        fit = LSQFitter(self.multispectrum, self.model, use_bounds=True)

        # if the start parameters are not yet calculated do it before the
        # model fitting
        if self.start_parameters is None:
            print('Estimates the parameters for the fitting procedure')
            self.estimate_start_parameters()

        fit.multi_fit(start_param=self.start_parameters)
        self.fitter = fit

        self.multispectrum.exclude = prev_exclude
        self.shift = fit.coeff_matrix[:, :, 1]

    def perform_alignment(self):
        """
        Performs the alignment procedure by first calculating the shift and
        then applying
        it to the spectra.

        Returns
        -------
        None.

        """
        self.calculate_model()
        self.align()



"""
Definition of the physical constants used with its units.
"""

import numpy as np


def R():
    """
    Rydberg constant in eV
    """

    return 13.606


def e():
    """
    Electron charge in C
    """

    return 1.602e-19


def m0():
    """
    Rest mass in kg
    """

    return 9.110e-31


def a0():
    """
    Bohr radius in m
    """

    return 5.292e-11


def speed_electron(E0):
    """
    Relativistic speed of electron

    """
    return c() * np.sqrt(1 - (1 / (1 + E0 * e() / (m0() * c() ** 2)) ** 2))


def joule_to_eV(J):
    return J * 6.24150913e18


def c():
    """
    Speed of light in m/s
    """

    return 2.998e8


def gamma(E0, m=m0()):
    """
    The formula works for energies of the electron
    which are not to close to the speed of light?

    :param E0: Energy incoming electron (eV)
        m: mass of the particle (kg)
        electron mass as default
    :return: gamma: float
    """
    gamma = 1 + e() * E0 / (m * c() ** 2)
    return gamma


def T(E0, m=m0()):
    """
    The effective incident energy in eV

    :param E0:energy in eV?
    :param m:mass in kg
    :return:
    """
    gam = gamma(E0, m=m)

    return E0 * (1 + gam) / (2 * gam ** 2)


def characteristic_angle(E, E0, use_rel=True):
    if use_rel:
        # theta_E = E / ((E0 + m0()*c()**2) *(speed_electron(E0)/c())**2)
        # theta_E = E / gamma(E0)*m0()*(speed_electron(E0)**2)
        theta_E = E / (2 * gamma(E0) * T(E0))
    else:
        theta_E = E / (2 * E0)
    return theta_E


def Zs_k(Z):
    """
    Corrects for the screening of the electrons
    for the k shell.
    :param Z: The number of protons
    :return: Corrected Z value
    """
    # return Z-0.3125 #value of EELSmodel
    return Z - 0.5  # value of hyperspy (Sigma3K)


def h(unit='J'):
    if unit == 'J':
        return 6.625607015e-34  # J/s
    elif unit == 'eV':
        return 4.135667696e-16  # eV/s


def hbar(unit='J'):
    return h(unit=unit) / (2 * np.pi)
